/* FILE: display.cc                 -*-Mode: c++-*-
 *
 * OOMMF colormap C++ code
 *
 */

#include <algorithm>
#include <string>
#include <vector>

#include "colormap.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

using SHADEVEC_T = Vf_DisplayVector::SHADEVEC_T;
using COLORSTR = ColorMapNS::COLORSTR;
using ColorMapNS::TKCOLORLENGTH;
using COLOR_MAP_FACTORY = ColorMap::COLOR_MAP_FACTORY;


//////////////////////////////////////////////////////////////////////////
// Utility Hue-Saturation-Value to Red-Green-Blue conversion routine,
// based on email from John Hagedorn (john.hagedorn@nist.gov), NIST x4339,
// on 3-Oct-1995.  All imports and exports in the range [0,1].
static void HSVtoRGB(double H,double S,double V,
                     double& R,double& G,double& B)
{
  H-=floor(H);  // Take hue input mod 1
  if(S>1.) S=1.; else if (S<0.) S=0.;
  if(V>1.) V=1.; else if (V<0.) V=0.;
  if(V==0) { R=G=B=0; return; }
  if(S==0) { R=G=B=V; return; }
  int sextant;  double fraction,p,q,t;
  sextant=(int)floor(H*6); fraction=H*6-sextant;
  p=V*(1.-S);  q=V*(1.-S*fraction);  t=V*(1.-S*(1.-fraction));
  switch(sextant) {
  default: // Safety
  case 0: // Red     -> Yellow
    R=V; G=t; B=p; break;
  case 1: // Yellow  -> Green
    R=q; G=V; B=p; break;
  case 2: // Green   -> Cyan
    R=p; G=V; B=t; break;
  case 3: // Cyan    -> Blue
    R=p; G=q; B=V; break;
  case 4: // Blue    -> Magenta
    R=t; G=p; B=V; break;
  case 5: // Magenta -> Red
    R=V; G=p; B=q; break;
  }
}

//////////////////////////////////////////////////////////////////////////
// OommfPackedRGB class
void OommfPackedRGB::Blend(OC_REAL8m wgt1,OommfPackedRGB color1,
			   OC_REAL8m wgt2,OommfPackedRGB color2)
{ // Generalization of Set; does wgt1*color1 + wgt2*color2.
  double newred=wgt1*double(color1.red)+wgt2*double(color2.red);
  if(newred<0.5)           red=0;
  else if(newred>=254.5)   red=255;
  else                     red=(unsigned char)OC_ROUND(newred);
  double newgreen=wgt1*double(color1.green)+wgt2*double(color2.green);
  if(newgreen<0.5)         green=0;
  else if(newgreen>=254.5) green=255;
  else                     green=(unsigned char)OC_ROUND(newgreen);
  double newblue=wgt1*double(color1.blue)+wgt2*double(color2.blue);
  if(newblue<0.5)          blue=0;
  else if(newblue>=254.5)  blue=255;
  else                     blue=(unsigned char)OC_ROUND(newblue);
}

OC_BOOL OommfPackedRGB::Set(const char* color_req)
{ // Converts color_req to rgb format, using Nb_GetColor.
  // Note that Nb_GetColor may call into the global Tcl interp.
  OC_REAL8m r,g,b;
  if(!Nb_GetColor(color_req,r,g,b)) return 0;
  if(r>=1.0) red=255;
  else       red=static_cast<unsigned char>(floor(r*256.));
  if(g>=1.0) green=255;
  else       green=static_cast<unsigned char>(floor(g*256.));
  if(b>=1.0) blue=255;
  else       blue=static_cast<unsigned char>(floor(b*256.));
  return 1;
}


////////////////////////////////////////////////////////////////////////
// Generic color maps
const ClassDoc ColorMap::class_doc("ColorMap",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","25-Mar-2024");

ColorMap* ColorMap::GetNewMap(const char* mapname,int colorcount) {
#define MEMBERNAME "GetNewMap(const char*,int)"
  // NB: Throws an Oc_Exception if mapname not recognized.
  ColorMap* newmap = nullptr;
  try {
    newmap = RegMaps().at(mapname).second();
    newmap->Setup(colorcount);
  } catch(std::out_of_range&) {
    // Match not found
    if(mapname == nullptr || mapname[0] == '\0') {
      // Return dummy colormap
      newmap = new NullColorMap();
    } else {
      std::string msg = "Unrecognized colormap request: ";
      msg += mapname;
      msg += "\nKnown maps ---\n";
      std::vector<std::string> mapnames;
      ColorMapList(mapnames);
      for(auto it : mapnames) { msg += it + "\n"; }
      msg += "---";
      OC_THROWEXCEPT(class_doc.classname,MEMBERNAME,msg);
    }
  }
  return newmap;
#undef MEMBERNAME
}

void ColorMap::ColorMapList
(std::vector<std::string>& colorlist)
{
  std::vector< std::pair<int,std::string> > namepairs;
  namepairs.reserve(RegMaps().size());
  for(const auto& kv : RegMaps()) {
    // kv.first is the mapname
    // kv.second is a pair of {id,factory function}
    // Use this to create a list of pairs {id, mapname}
    namepairs.push_back(std::make_pair(kv.second.first,kv.first));
  }
  // Sort namepairs by mapid
  std::sort(namepairs.begin(), namepairs.end());

  // Copy mapnames from sorted list to colorlist export
  colorlist.clear();
  colorlist.reserve(namepairs.size());
  for(const auto& kv : namepairs) {
    colorlist.push_back(kv.second);
  }
}

template <typename T>
 ColorMap* MakeNewColorMap() {
 return static_cast<ColorMap*>(new T());
}

////////////////////////////////////////////////////////////////////////
// ColorMap registration
template <typename T>
class RegisterColorMap {
public:
  RegisterColorMap() {
    ColorMap::Register(T::GetClassName(),MakeNewColorMap<T>);
  }
};
// Usage example:
//    namespace { RegisterColorMap<ColorMapRedBlackBlue> dummy; }

////////////////////////////////////////////////////////////////////////
// 1D Color maps
class ColorMap1D : public ColorMap
{
public:
  ColorMap1D() = default;
  ColorMap1D(const ColorMap1D&) = default;
  ColorMap1D& operator=(const ColorMap1D&) = default;
  virtual ~ColorMap1D() = default;

  /// NB: The Clone member must be defined in the each leaf class so
  /// that the ColorMap* resolves to the leaf class.

  size_t GetColorCount() const { return colors.size(); }

  COLORSTR GetColor(const Vf_DisplayVector::SHADEVEC_T& shade) const {
    // Import shade[0] hould in [0.,1.]
    const size_t colorcount = colors.size();
    double chkval = OC_ROUND(shade[0]*double(colorcount-1));
    size_t index = 0;
    if(chkval > 0.0) {
      if (chkval <  static_cast<double>(colorcount)) {
        index = static_cast<size_t>(chkval);
      } else {
        index = colorcount - 1;
      }
    }
    return colors[index];
  }

protected:
  virtual void MakeColor(const SHADEVEC_T& relative_shade,
                         COLORSTR& color) = 0;

private:
  size_t Setup(size_t req_color_count) {
    size_t color_count = (req_color_count>1 ? req_color_count : 1);
    colors.clear();
    colors.reserve(color_count);
    COLORSTR tmpclr;
    if(color_count==1) {
      MakeColor(0.,tmpclr);
      colors.emplace(colors.end(),tmpclr);
    } else {
      for(size_t i=0;i<color_count;i++) {
        MakeColor(static_cast<OC_REAL4m>(double(i)/double(color_count-1)),
                  tmpclr);
        colors.emplace(colors.end(),tmpclr);
      }
    }
    return colors.size();
  }

  std::vector<COLORSTR> colors;
  /// Leave room for '#' and '\0'
};

// A few sample ColorMap1D classes.  Each of these map a linear
// spectrum through an arccos, which will display a component map
// as linear in angle.
class ColorMapRedBlackBlue : public ColorMap1D
{
public:
  static std::string GetClassName() { // Static mapname for registration
    return std::string("Red-Black-Blue");
  }
  virtual std::string GetMapName() const {
    return GetClassName(); // Virtual call for instance mapname
  }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapRedBlackBlue(*this);
  }

  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    int red,green,blue;
    red=blue=0; // To keep dumb compilers quiet
    green=0;
    if(value<128) {
      blue=0;
      red=255-2*value;
    } else if(value>=128) {
      red=0;
      blue=2*(value-128)+1;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapRedBlackBlue> dummy0; }


class ColorMapBlueWhiteRed : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Blue-White-Red");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlueWhiteRed(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    int red,green,blue;
    red=green=blue=0; // To keep dumb compilers quiet
    if(value<128) {
      blue=255;
      red=green=2*value;
    } else if(value>=128) {
      red=255;
      blue=green=510-2*value;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapBlueWhiteRed> dummy1; }


#if 0
/// 1D ColorMap Template:
class ColorMapBlueWhiteRed : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Blue-White-Red");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlueWhiteRed(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    // Map specific code goes here
  }
};
namespace { RegisterColorMap<ColorMapBlueWhiteRed> dummy; }
#endif // 0

class ColorMapTealWhiteRed : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Teal-White-Red");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapTealWhiteRed(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    int red,green,blue;
    red=green=blue=0; // To keep dumb compilers quiet
    if(value<128) {
      blue=255;
      green=128+value;
      red=2*value;
    } else if(value>=128) {
      red=255;
      blue=green=510-2*value;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapTealWhiteRed> dummy2; }

class ColorMapGreenWhiteOrange : public ColorMap1D
{ // Erin Go Braugh
public:
  static std::string GetClassName() {
    return std::string("Green-White-Orange");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapGreenWhiteOrange(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    int red,green,blue;
    red=green=blue=0; // To keep dumb compilers quiet
    if(value<128) {
      green=191+value/2;
      red=blue=2*value;
    } else if(value>=128) {
      red=255;
      blue=510-2*value;
      green=345-(45*value)/64;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapGreenWhiteOrange> dummy3; }

class ColorMapBlackGrayWhite : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Black-Gray-White");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlackGrayWhite(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapBlackGrayWhite> dummy4; }

class ColorMapHighContrastGray : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("HighContrastGray");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapHighContrastGray(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    const double contrast=2.0;
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x = relative_shade[0];
      x = atan(contrast*(2*x-1)/(x*(1-x)))/PI+0.5;
      value=(int)floor(256*x);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapHighContrastGray> dummy5; }

class ColorMapBlackWhiteBlack : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Black-White-Black");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlackWhiteBlack(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 0;
    else if(relative_shade[0]>0.0) {
      double x=acos(fabs(1-2*relative_shade[0]));
      value=(int)floor(512*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapBlackWhiteBlack> dummy6; }

class ColorMapWhiteBlackWhite : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("White-Black-White");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapWhiteBlackWhite(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 0;
    else if(relative_shade[0]>0.0) {
      double x=acos(fabs(1-2*relative_shade[0]));
      value=(int)floor(512*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    value = 255 - value;
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapWhiteBlackWhite> dummy7; }

class ColorMapBlackWhiteWhite : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Black-White-White");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlackWhiteWhite(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(512*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapBlackWhiteWhite> dummy8; }

class ColorMapBlackBlackWhite : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Black-Black-White");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapBlackBlackWhite(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(512*x/PI)-256;
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",value,value,value);
  }
};
namespace { RegisterColorMap<ColorMapBlackBlackWhite> dummy9; }

class ColorMapWhiteGreenBlack : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("White-Green-Black");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapWhiteGreenBlack(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    int value=0;
    if(relative_shade[0]>=1.0) value = 255;
    else if(relative_shade[0]>0.0) {
      double x=acos(1-2*relative_shade[0]);
      value=(int)floor(256*x/PI);
      if(value<0)   value=0;
      if(value>255) value=255;
    }
    int red,green,blue;
    red=blue=255-value;
    green=255;
    if(value>127) green=510-2*value;
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapWhiteGreenBlack> dummy10; }

class ColorMapRedGreenBlueRed : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Red-Green-Blue-Red");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapRedGreenBlueRed(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    double R,G,B;
    HSVtoRGB(relative_shade[0],1.,1.,R,G,B);
    int   red = int(OC_ROUND(R*256.));  if(  red>255)   red=255;
    int green = int(OC_ROUND(G*256.));  if(green>255) green=255;
    int  blue = int(OC_ROUND(B*256.));  if( blue>255)  blue=255;
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
// Registration grouped at bottom of file with other circular maps

class ColorMapCyanYellowMagentaCyan : public ColorMap1D
{
public:
  static std::string GetClassName() {
    return std::string("Cyan-Yellow-Magenta-Cyan");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapCyanYellowMagentaCyan(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, COLORSTR& color) {
    SHADEVEC_T shifted_shade(relative_shade);
    shifted_shade[0] -= 0.5;
    if(shifted_shade[0]<0) shifted_shade[0] += 1.0;
    double R,G,B;
    HSVtoRGB(shifted_shade[0],1.,1.,R,B,G); // Reflect green and blue
    int   red = int(OC_ROUND(R*256.));  if(  red>255)   red=255;
    int green = int(OC_ROUND(G*256.));  if(green>255) green=255;
    int  blue = int(OC_ROUND(B*256.));  if( blue>255)  blue=255;
    snprintf(color.data(),sizeof(color),"#%02X%02X%02X",red,green,blue);
  }
};
// Registration grouped at bottom of file with other circular maps

////////////////////////////////////////////////////////////////////////
// 2D Color maps
#define USE_OLD_PLUS_SAT 0
#if USE_OLD_PLUS_SAT
// First pass at hue colormap with saturation shading. This uses a
// simple approximation to saturation, as compared to the
// ColorMapHuePlusSat class. This class and children are disabled by
// default, but may bear revisiting if ColorMapHuePlusSat is too slow.
class ColorMap1DPlusSat : public ColorMap
{ // This class is based on ColorMap1D, but with color saturation
  // affected by shade[1].

  // Note 1: The internal colormap in this class ("colors") is an
  //   OommfPackedRGB array rather than an array of Tk #rrggbb strings.
  // Note 2: The color quantization value ColorCount is the number of
  //   hues. The RGB color is computed on request by layering in the
  //   saturation, so the number of actual distinct RGB color outputs
  //   is potentially ColorCount * (# of floating point values
  //   in [0,1]), where the fp type is decltype(shade[1]).
public:
  ColorMap1DPlusSat() = default;
  ColorMap1DPlusSat(const ColorMap1DPlusSat&) = default;
  ColorMap1DPlusSat& operator=(const ColorMap1DPlusSat&) = default;
  virtual ~ColorMap1DPlusSat() = default;

  /// NB: The Clone member must be defined in the each leaf class so
  /// that the ColorMap* resolves to the leaf class.

  size_t GetColorCount() const { return colors.size(); }

  COLORSTR GetColor(const Vf_DisplayVector::SHADEVEC_T& shade) const {
    // Imports shade[0] and shade[1] should in [0.,1.]
    const size_t colorcount = colors.size();
    int index=(int)OC_ROUND(shade[0]*double(colorcount-1));
    if(index<0) {
      index=0;
    } else if(static_cast<size_t>(index) >= colorcount) {
      index=colorcount-1;
    }

    // Saturation adjustment
    // int adjustment = (int)OC_ROUND(255*sqrt(2*fabs(shade[1]-0.5)));
    int adjustment = (int)OC_ROUND(255*(2*fabs(shade[1]-0.5)));
    unsigned int r,g,b;
    colors[index].Get(r,g,b);
    if((r+=adjustment) > 255) r = 255;
    if((g+=adjustment) > 255) g = 255;
    if((b+=adjustment) > 255) b = 255;
    OommfPackedRGB tweaked_color(r,g,b);
    COLORSTR tkcolor;
    tweaked_color.Get(tkcolor);
    return tkcolor;
  }

protected:
  virtual void MakeColor(const SHADEVEC_T& relative_shade,
                         OommfPackedRGB& color) = 0;

private:
  size_t Setup(size_t req_color_count) {
    size_t color_count = (req_color_count>1 ? req_color_count : 1);
    colors.clear();
    colors.reserve(color_count);
    OommfPackedRGB tmpclr;
    if(color_count==1) {
      MakeColor(0.,tmpclr);
      colors.emplace(colors.end(),tmpclr);
    } else {
      for(size_t i=0;i<color_count;i++) {
        MakeColor(double(i)/double(color_count-1),tmpclr);
        colors.emplace(colors.end(),tmpclr);
      }
    }
    return colors.size();
  }

  std::vector<OommfPackedRGB> colors;
};

class ColorMapRedGreenBlueRedPlusSatSimple : public ColorMap1DPlusSat
{
public:
  static std::string GetClassName() {
    return std::string("Red-Green-Blue-Red+SatSimple");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapRedGreenBlueRedPlusSatSimple(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, OommfPackedRGB& color) {
    double R,G,B;
    HSVtoRGB(relative_shade[0],1.,1.,R,G,B);
    unsigned int   red = static_cast<unsigned int>(OC_ROUND(R*256.));
    if(  red>255)   red=255;
    unsigned int green = static_cast<unsigned int>(OC_ROUND(G*256.));
    if(green>255) green=255;
    unsigned int  blue = static_cast<unsigned int>(OC_ROUND(B*256.));
    if( blue>255)  blue=255;
    color.Set(red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapRedGreenBlueRedPlusSatSimple> dummy100; }

class ColorMapCyanYellowMagentaCyanPlusSatSimple : public ColorMap1DPlusSat
{
public:
  static std::string GetClassName() {
    return std::string("Cyan-Yellow-Magenta-Cyan+SatSimple");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapCyanYellowMagentaCyanPlusSatSimple(*this);
  }
  void MakeColor(const SHADEVEC_T& relative_shade, OommfPackedRGB& color) {
    SHADEVEC_T shifted_shade(relative_shade);
    shifted_shade[0] -= 0.5;
    if(shifted_shade[0]<0) shifted_shade[0] += 1.0;
    double R,G,B;
    HSVtoRGB(shifted_shade[0],1.,1.,R,B,G); // Reflect green and blue
    unsigned int   red = static_cast<unsigned int>(OC_ROUND(R*256.));
    if(  red>255)   red=255;
    unsigned int green = static_cast<unsigned int>(OC_ROUND(G*256.));
    if(green>255) green=255;
    unsigned int  blue = static_cast<unsigned int>(OC_ROUND(B*256.));
    if( blue>255)  blue=255;
    color.Set(red,green,blue);
  }
};
namespace { RegisterColorMap<ColorMapCyanYellowMagentaCyanPlusSatSimple> dummy101; }
#endif // USE_OLD_PLUS_SAT

////////////////////////////////////////////////////////////////////////
// 2D Hue+Sat colormap
class ColorMapHuePlusSat : public ColorMap
{ // Conceptually similar to ColorMap1DPlusSat, but using hue and
  // saturation in the HSV model, with value=1. All values are
  // computed on the fly, i.e., there is no color lookup table.  The
  // color quantization value ColorCount is the number of hues. The
  // RGB color is computed on request by layering in the saturation,
  // so the number of actual distinct RGB color outputs is potentially
  // ColorCount * (# of floating point values in [0,1]), where the fp
  // type is decltype(shade[1]).

public:
  ColorMapHuePlusSat() = default;
  ColorMapHuePlusSat(const ColorMapHuePlusSat&) = default;
  ColorMapHuePlusSat& operator=(const ColorMapHuePlusSat&) = default;
  virtual ~ColorMapHuePlusSat() = default;

  /// NB: The Clone member must be defined in the each leaf class so
  /// that the ColorMap* resolves to the leaf class.

  size_t GetColorCount() const { return color_count; }

  COLORSTR GetColor(const Vf_DisplayVector::SHADEVEC_T& shade) const {
    // Imports shade[0] and shade[1] should in [0.,1.]
    const double saturation = 1.0 - 2*fabs(shade[1]-0.5);
    const double value      = 1.0;
    double hue = (invert ? 1-shade[0] : shade[0]) + coloroffset;
    hue -= floor(hue);
    /// Note: Here hue is always in [0,1), but the hue map is circular
    /// so hue == 0 is the same as hue == 1.
    if(color_count>0) { // Discretize hue
      hue = floor(color_count*hue + 0.5)/double(color_count);
    }
    double r,g,b;
    HSVtoRGB(hue,saturation,value,r,g,b);
    assert(TKCOLORLENGTH == 6);
    COLORSTR tkcolor;
    snprintf(tkcolor.data(),sizeof(tkcolor),"#%02X%02X%02X",
             static_cast<int>(floor(255*r+0.5)),
             static_cast<int>(floor(255*g+0.5)),
             static_cast<int>(floor(255*b+0.5)));
    /// snprintf() might be slow. May want to investigate a quicker
    /// way to make this conversion.
    return tkcolor;
  }

protected:
  // Offset and Invert dictate hue rotation.
  virtual bool Invert() const =0;  // Invert color ordering?
  virtual double ColorOffset() const =0; // Child specifies offset
  /// ColorOffset return should be in the range [0,1).

private:
  size_t Setup(size_t req_color_count) {
    color_count = (req_color_count>0 ? req_color_count : 0);
    /// 0 means no color discretization
    invert = this->Invert();
    coloroffset = this->ColorOffset();
    return color_count;
  }
  size_t color_count;
  bool invert;
  double coloroffset;
};

class ColorMapRedGreenBlueRedPlusSat : public ColorMapHuePlusSat
{
public:
  static std::string GetClassName() {
    return std::string("Red-Green-Blue-Red+Sat");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapRedGreenBlueRedPlusSat(*this);
  }
  bool  Invert() const { return 0; }  // Don't invert color ordering
  double ColorOffset() const { return 0.0; } // Red-Green-Blue-Red == 0 offset
};
// Registration grouped at bottom of file with other circular maps

class ColorMapCyanYellowMagentaCyanPlusSat : public ColorMapHuePlusSat
{ // This map is the same as phase shifting ColorMapRedGreenBlueRedPlusSat
  // by 180 degrees; it is provided as a convenience for the mmDisp UI.
public:
  static std::string GetClassName() {
    return std::string("Cyan-Yellow-Magenta-Cyan+Sat");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapCyanYellowMagentaCyanPlusSat(*this);
  }
  bool  Invert() const { return 1; }  // Invert color ordering
  double ColorOffset() const { return (1./2.); }
  // Cyan-Yellow-Magenta-Cyan is 180 degrees offset from RGBR
  // => offset=180/360
};
// Registration grouped at bottom of file with other circular maps


////////////////////////////////////////////////////////////////////////
// 2D Hue+Sat/Val colormap
class ColorMapHuePlusSatVal : public ColorMap
{ // Conceptually similar to ColorMapHuePlusSat, and in fact exactly
  // the same as ColorMapHuePlusSat for shade[1]>0.5.  (In this region
  // value is fixed at 1.0.)  But for shade[1]<0.5 saturation is fixed
  // at 1.0 and value is decreased from 1.0 at shade[1]==0.5 to 0.0 at
  // shade[1]==0.0. The first take tried the linear mapping
  //       value = 1.0 - 2*(0.5 - shade[1]);
  // but this appears to darken too fast, so the second try is
  //       value = sqrt(1.0 - 2*(0.5 - shade[1]));
  // which seems to work better.
  //
  // All values are computed on the fly, i.e., there is no color
  // lookup table.  The color quantization value ColorCount is the
  // number of hues. The RGB color is computed on request by layering
  // in the saturation and value, so the number of actual distinct RGB
  // color outputs is potentially ColorCount * 2 * (# of floating
  // point values in [0,1]), where the fp type is decltype(shade[1]).
  //
  // When combining both arrow and pixel colormaps, one might try
  // for example plot .config settings like
  //
  //   arrow,colorcount          2
  //   arrow,colormap            Black-Gray-White
  //   arrow,colorreverse        1
  //   arrow,quantity            x
  //   ...
  //   pixel,colorcount          0
  //   pixel,colormap            Red-Green-Blue-Red+SatVal
  //   pixel,quantity            yzangle
  //
  // This colors the arrow black on the shade[1]>0.5 areas where
  // value=1, 0<saturation<1, and white where shade[1]<0.5 where
  // saturation=0, 0<value<1.
  //
public:
  ColorMapHuePlusSatVal() = default;
  ColorMapHuePlusSatVal(const ColorMapHuePlusSatVal&) = default;
  ColorMapHuePlusSatVal& operator=(const ColorMapHuePlusSatVal&) = default;
  virtual ~ColorMapHuePlusSatVal() = default;

  /// NB: The Clone member must be defined in the each leaf class so
  /// that the ColorMap* resolves to the leaf class.

  size_t GetColorCount() const { return color_count; }

  COLORSTR GetColor(const Vf_DisplayVector::SHADEVEC_T& shade) const {
    // Imports shade[0] and shade[1] should in [0.,1.]
    double saturation, value;
    if(shade[1]>0.5) {
      // Same as ColorMapRedGreenBlueRedPlusSat
      saturation = 1.0 - 2*(shade[1]-0.5);
      value = 1.0;
    } else {
      // Similar to above, but swapping saturation and value
      value = 1.0 - 2*(0.5 - shade[1]);
      value = sqrt(value); // Feather value darkening
      saturation = 1.0;
    }
    double hue = (invert ? 1-shade[0] : shade[0]) + coloroffset;
    hue -= floor(hue);
    /// Note: Here hue is always in [0,1), but the hue map is circular
    /// so hue == 0 is the same as hue == 1.
    if(color_count>0) { // Discretize hue
      hue = floor(color_count*hue + 0.5)/double(color_count);
    }
    double r,g,b;
    HSVtoRGB(hue,saturation,value,r,g,b);
    assert(TKCOLORLENGTH == 6);
    COLORSTR tkcolor;
    snprintf(tkcolor.data(),sizeof(tkcolor),"#%02X%02X%02X",
             static_cast<int>(floor(255*r+0.5)),
             static_cast<int>(floor(255*g+0.5)),
             static_cast<int>(floor(255*b+0.5)));
    /// snprintf() might be slow. May want to investigate a quicker
    /// way to make this conversion.
    return tkcolor;
  }

protected:
  // ColorOffset and Invert dictate hue rotation.
  virtual bool  Invert() const =0; // Invert color ordering?
  virtual double ColorOffset() const =0; // Child specifies offset.
  /// Offset should be in the range [0,1).

private:
  size_t Setup(size_t req_color_count) {
    color_count = (req_color_count>0 ? req_color_count : 0);
    /// 0 means no color discretization
    invert = this->Invert();
    coloroffset = this->ColorOffset();
    return color_count;
  }
  size_t color_count;
  bool invert;
  double coloroffset;
};

class ColorMapRedGreenBlueRedPlusSatVal : public ColorMapHuePlusSatVal
{
public:
  static std::string GetClassName() {
    return std::string("Red-Green-Blue-Red+SatVal");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapRedGreenBlueRedPlusSatVal(*this);
  }
  bool  Invert() const { return 0; }  // Don't invert color ordering
  double ColorOffset() const { return 0.0; } // Red-Green-Blue-Red == 0 offset
};
// Registration grouped at bottom of file with other circular maps

class ColorMapCyanYellowMagentaCyanPlusSatVal : public ColorMapHuePlusSatVal
{ // This map is the same as phase shifting ColorMapRedGreenBlueRedPlusSatVal
  // by 180 degrees; it is provided as a convenience for the mmDisp UI.
public:
  static std::string GetClassName() {
    return std::string("Cyan-Yellow-Magenta-Cyan+SatVal");
  }
  virtual std::string GetMapName() const { return GetClassName(); }
  virtual ColorMap* Clone() { // Allows deep copies from base class
    return new ColorMapCyanYellowMagentaCyanPlusSatVal(*this);
  }
  bool  Invert() const { return 1; }  // Invert color ordering
  double ColorOffset() const { return (1./2.); }
  // Cyan-Yellow-Magenta-Cyan is 180 degrees offset from RGBR
  // => offset=180/360
};
// Registration grouped at bottom of file with other circular maps


// Registration for circular maps. ColorMap::ColorMapList() orders
// the map name list by registration order.
namespace { RegisterColorMap<ColorMapRedGreenBlueRed> dummy11; }
namespace { RegisterColorMap<ColorMapRedGreenBlueRedPlusSat> dummy200; }
namespace { RegisterColorMap<ColorMapRedGreenBlueRedPlusSatVal> dummy300; }
namespace { RegisterColorMap<ColorMapCyanYellowMagentaCyan> dummy12; }
namespace { RegisterColorMap<ColorMapCyanYellowMagentaCyanPlusSat> dummy201; }
namespace { RegisterColorMap<ColorMapCyanYellowMagentaCyanPlusSatVal> dummy301; }


//////////////////////////////////////////////////////////////////////////
// DisplayColorMap class
const ClassDoc DisplayColorMap::class_doc("DisplayColorMap",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "2.0.0","25-Mar-2024");
