/* FILE: colormap.h                 -*-Mode: c++-*-
 *
 * OOMMF colormap C++ class
 *
 */

#ifndef COLORMAP_H
#define COLORMAP_H

#include <stdexcept>
#include <functional> // std::function
#include <memory>     // std::unique_ptr
#include <unordered_map>
#include <utility>    // std::pair

#include "nb.h"
#include "vf.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// Color specifications
//

namespace ColorMapNS {
  // Number of hex characters in #rrggbb
  static constexpr size_t TKCOLORLENGTH = 6;
  using COLORSTR = std::array<char,TKCOLORLENGTH+2>; // +2 for # and '\0'
}

class OommfPackedRGB {
protected:
  // If the following variables are made private rather than protected,
  // then the Intel C++ compiler warns that
  //     private field 'pad' is not used
  // Workarounds are (a) drop pad, (b) silence warning, or (c) make
  // 'pad' protected. Tentatively choosing (c) for now. YMMV.
  unsigned char red,green,blue,pad;
public:
  OommfPackedRGB() : red(0), green(0), blue(0), pad(0) {}
  OommfPackedRGB(const char* tkcolorstr) {
    Set(tkcolorstr);
  }
  OommfPackedRGB(int r,int g,int b)
    : red((unsigned char)r), green((unsigned char)g),
      blue((unsigned char)b), pad(0) {}

  void Get(unsigned int &myred,unsigned int &mygreen,
	   unsigned int &myblue) const {
    myred=red; mygreen=green; myblue=blue;
  }

  void Get(ColorMapNS::COLORSTR& tkcolor) const {
    assert(ColorMapNS::TKCOLORLENGTH == 6);
#if 0 // Bog standard
    snprintf(tkcolor.data(),sizeof(tkcolor),
             "#%02X%02X%02X",red,green,blue);
#else // Bog++
    // Testing indicates that this version is 30-80 times faster than
    // using snprintf, depending on the hardware. (One pass here
    // typicaly takes 1-5 ns, vs 70-150 ns w/ snprintf. Cf. test
    // program oommf/dvl/colorconvert-speedtest.cc.)
    tkcolor[0]='#';
    std::array<char,2> rbuf(ConvertIntCharToHexByte(red));
    tkcolor[1] = rbuf[0]; tkcolor[2] = rbuf[1];
    std::array<char,2> gbuf(ConvertIntCharToHexByte(green));
    tkcolor[3] = gbuf[0]; tkcolor[4] = gbuf[1];
    std::array<char,2> bbuf(ConvertIntCharToHexByte(blue));
    tkcolor[5] = bbuf[0]; tkcolor[6] = bbuf[1];
    tkcolor[7] = '\0';
#endif // Boggish
  }

  OC_BOOL Set(const char* color_req);
  /// Uses Nb_GetColor to convert color_req into an rgb triplet.  If
  /// color_req is in Tk numeric format, e.g. #FFFFFF, then Nb_GetColor
  /// does the conversion directly.  Otherwise, Nb_GetColor uses the
  /// global Tcl interpreter to access the nbcolordatabase array defined
  /// in oommf/config/colors.config.

  OC_BOOL Set(unsigned int new_red,unsigned int new_green,
	   unsigned int new_blue) {
    red   = (unsigned char)new_red;
    green = (unsigned char)new_green;
    blue  = (unsigned char)new_blue;
    return 1;
  }

  void Blend(OC_REAL8m wgt1,OommfPackedRGB color1,
	     OC_REAL8m wgt2,OommfPackedRGB color2);
  /// Generalization of Set, using wgt1*color1 + wgt2*color2.

  void Invert() {
    red   = (unsigned char)(255-red);
    green = (unsigned char)(255-green);
    blue  = (unsigned char)(255-blue);
  }

  static std::array<char,2> ConvertIntCharToHexByte(unsigned char ic) {
    // Import ic is an unsigned character value in range [0,255]. The
    // return is two hexadecimal characters; Example ic = 43 = 0x2B maps
    // to '2B'. Testing indicates that this code runs 2-10 times faster
    // than converting through strtol, depending on the hardware.  (One
    // pass typically takes 1-8 ns with this code, vs. 8-20 ns with
    // strtol. Cf. test program oommf/dvl/colorconvert-speedtest.cc.)
    static IntCharToHexByte icthb;
    return icthb.map[ic];
  }

  static int ConvertHex2ToInt(const std::array<char,2>& hh) {
    // Import hh is two hex characters (0-9 and either [a-f] or
    // [A-F]). The return is an integer in the range 0-255. {'2','B'}
    // => 0x2B = 43. Note: I tried using a std::unordered_map to do
    // this, but std::unordered_map requires a hash function for the
    // key, and C++11 does not have a built-in hash for std::array.
    // So instead use a simple prefilled array on each digit.
    static HexToInt hti;
    assert(hti.arrmap[hh[0]]>=0 && hti.arrmap[hh[1]]>=0);
    return 16*hti.arrmap[hh[0]] + hti.arrmap[hh[1]];
  }

private:
  // Array mapping one byte integer value to two hex digits ("%02X").
  // This is a support struct for
  //    ConvertIntCharToHexByte(unsigned char)
  struct IntCharToHexByte {
  public:
    IntCharToHexByte() {
      char buf[3];
      for(unsigned int i=0;i<map.size();++i) {
        snprintf(buf,sizeof(buf),"%02X",i);
        map[i][0] = buf[0];
        map[i][1] = buf[1];
      }
    }
    std::array< std::array<char,2>, 256> map;
  };

  // Array mapping single hex digit (0-9, A-F, a-f) to integer value.
  // This is a support struct for
  //    ConvertHex2ToInt(const std::array<char,2>&)
  struct HexToInt {
  public:
    HexToInt() {
      arrmap.fill(-1); // Fill for error detection
      char buf[2];
      for(int i=0;i<16;++i) {
        snprintf(buf,sizeof(buf),"%X",i);
        arrmap[buf[0]] = i;
        if(i>9) {
          snprintf(buf,sizeof(buf),"%x",i);
          arrmap[buf[0]] = i;
        }
      }
    }
    std::array<int,256> arrmap;
  };

};

//////////////////////////////////////////////////////////////////////////
// Generic ColorMap class
class ColorMap
{
private:
  static const ClassDoc class_doc;
public:
  using COLOR_MAP_FACTORY = std::function<ColorMap*()>;
  using SHADEVEC_T = Vf_DisplayVector::SHADEVEC_T;

  ColorMap() = default;
  ColorMap(const ColorMap&) = default;
  ColorMap& operator=(const ColorMap&) = default;

  virtual ~ColorMap() = default;

  virtual ColorMap* Clone() =0;

  virtual size_t Setup(size_t colorcount) =0;

  static ColorMap* GetNewMap(const char* mapname,int colorcount);

  virtual size_t GetColorCount() const =0;
  virtual std::string GetMapName() const =0;
  virtual ColorMapNS::COLORSTR
     GetColor(const Vf_DisplayVector::SHADEVEC_T& shade) const =0;
  OommfPackedRGB
  GetPackedColor(const Vf_DisplayVector::SHADEVEC_T& shade) const {
    const ColorMapNS::COLORSTR tmpclr = GetColor(shade);
    assert(ColorMapNS::TKCOLORLENGTH == 6);
    int red   = OommfPackedRGB::ConvertHex2ToInt({tmpclr[1],tmpclr[2]});
    int green = OommfPackedRGB::ConvertHex2ToInt({tmpclr[3],tmpclr[4]});
    int blue  = OommfPackedRGB::ConvertHex2ToInt({tmpclr[5],tmpclr[6]});
    return {red,green,blue};
  }

  static void Register(std::string _leafname,
                       COLOR_MAP_FACTORY _colormap) {
    static int id = 0;
    RegMaps()[_leafname] = { id++, _colormap };
  }

  static void ColorMapList(std::vector<std::string>& mapnames);

private:
  // Note: The available colormaps are stored in an std::unordered_map
  // object. To work around static initialization issues, this map is
  // a static object inside the static accessor function RegMaps(); A
  // similar technique is used for Oxs_Ext objects in the OOMMF Oxs
  // application (see oommf/app/oxs/base/ext.cc; also Section 9.4.1 of
  // "The C++ Programming Lanuage," Bjarne Stroustrup, 3rd edition for
  // a discussion of nonlocal variable initialization order.)
  //   Each map name string is associated with a std::pair
  // consisting of the registration id (used for sorting) and
  // a pointer to the ColorMap factory function.
  static std::unordered_map<std::string,
                            std::pair<int,COLOR_MAP_FACTORY> >& RegMaps() {
    static std::unordered_map<std::string,
                              std::pair<int,COLOR_MAP_FACTORY> > maps;
    return maps;
  }
};

//////////////////
// NullColorMap
// This is a dummy ColorMap class that can be used workaround
// for pre-main() "static" initializations issues (e.g.,
// 'static DisplayFrame myFrame' in mmdispcmds.cc). This class
// does not call ColorMap::Register and is does not appear in
// ColorMap::ColorMapList().

class NullColorMap : public ColorMap
{
  virtual ColorMap* Clone() { return new NullColorMap(*this); }
  virtual size_t Setup(size_t) { return 0; }
  virtual size_t GetColorCount() const { return 0; }
  virtual std::string GetMapName() const { return std::string(""); }
  virtual ColorMapNS::COLORSTR
  GetColor(const Vf_DisplayVector::SHADEVEC_T&) const {
    return ColorMapNS::COLORSTR { '#','0','0','0','0','0','0','\0' };
    /// Black
  }
};

//////////////////////////////////////////////////////////////////////////
// DisplayColormap class
//
class DisplayColorMap {
private:
  static const ClassDoc class_doc;
  std::unique_ptr<ColorMap> colormap;
public:
  using SHADEVEC_T = Vf_DisplayVector::SHADEVEC_T;

  // The PlotConfiguration class holds a DisplayColorMap instance.
  // To support PlotConfiguration creation and copying we therefore
  // need DisplayColorMap(), DisplayColorMap(const DisplayColorMap&),
  // and DisplayColorMap& operator=(const DisplayColorMap&);

  void Setup(const char* mapname, int colorcount) {
    try {
      ColorMap* newmap = ColorMap::GetNewMap(mapname,colorcount);
      colormap.reset(newmap); // No change if error in GetNewMap()
    } catch(Oc_Exception& ocexcept) {
      NonFatalError(ocexcept);
    }
  }

  DisplayColorMap() : colormap(new NullColorMap()) {}
  /// Note: NullColorMap is a dummy placeholder to allow
  /// pre-main static initialization---see note on 'static
  /// DisplayFrame myFrame' in mmdispcmds.cc. The initial
  /// working colormap is set from the application .config
  /// file.
  DisplayColorMap(int colorcount,const char* mapname) {
    Setup(mapname,colorcount);
  }
  DisplayColorMap(const DisplayColorMap& other)
    : colormap(other.colormap->Clone()) // Deep copy
  {}

  DisplayColorMap& operator=(const DisplayColorMap& other) {
    colormap.reset(other.colormap->Clone()); // Deep copy
    return *this;
  }

  ~DisplayColorMap() = default;

  static void ColorMapList(std::vector<std::string>& maps) {
    ColorMap::ColorMapList(maps);
  }

  std::string GetMapName() { return colormap->GetMapName(); }
  size_t GetColorCount() { return colormap->GetColorCount(); }

  ColorMapNS::COLORSTR GetColor(const SHADEVEC_T& relative_shade) const {
    return colormap->GetColor(relative_shade);
  }
  OommfPackedRGB GetPackedColor(const SHADEVEC_T& relative_shade) const {
    return colormap->GetPackedColor(relative_shade);
  }
};

#endif /* COLORMAP_H */
