/* FILE: psdraw.cc                 -*-Mode: c++-*-
 *
 * OOMMF C++ classes for PostScript rendering
 *
 */

#include <ctime>
#include <sstream>
#include <utility>
#include <vector>

#include "psdraw.h"
#include "nb.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

////////////////////////////////////////////////////////////////////////
// Initialize OommfDrawVectorGraphics base object.  Offsets and print
// sizes are in points, px/y range values in pixels.  x/yoff are to
// lower lefthand corner of image.  Pixel coordinates run from upper
// lefthand corner of image.
//
// Child classes should perform their own initialization and then write
// their header to "channel".
OommfDrawVectorGraphics::OommfDrawVectorGraphics
(Tcl_Channel use_channel,
 OC_REAL8m pxoff,OC_REAL8m pyoff,
 OC_REAL8m printwidth,OC_REAL8m printheight,
 const char* pageorientation,
 OC_REAL8m xmin,OC_REAL8m ymin,
 OC_REAL8m xmax,OC_REAL8m ymax,
 OommfPackedRGB background_color,
 OC_REAL8m matwidth,
 OommfPackedRGB matcolor,
 OommfPackedRGB arrow_outline_color,
 OC_REAL8m arrow_outline_width,
 OC_REAL8m arrow_width,
 OC_REAL8m arrow_tip_width,
 OC_REAL8m arrow_length,
 OC_REAL8m arrow_head_width,
 OC_REAL8m arrow_head_ilen,
 OC_REAL8m arrow_head_olen)
  : channel(use_channel),
    bbox_xmin(int(floor(pxoff))),
    bbox_ymin(int(floor(pyoff))),
    bbox_xmax(int(ceil(pxoff
     +(strcmp("Landscape",pageorientation)==0 ? printheight : printwidth)))),
    bbox_ymax(int(ceil(pyoff
     +(strcmp("Landscape",pageorientation)==0 ? printwidth : printheight)))),
    scale(OC_MIN(printwidth/(xmax-xmin),printheight/(ymax-ymin))),
    mat_width(matwidth),
    mat_color(matcolor),
    ArrowOutlineColor(arrow_outline_color),
    ArrowOutlineWidth(arrow_outline_width),
    ArrowWidth(arrow_width),
    ArrowTipWidth(arrow_tip_width),
    ArrowLength(arrow_length),
    ArrowHeadWidth(arrow_head_width),
    ArrowHeadInnerLength(arrow_head_ilen),
    ArrowHeadOuterLength(arrow_head_olen),
    InOutTipRadius(0.15),
    page_orientation(strcmp("Landscape",pageorientation)==0
                     ? PageOrientation::Landscape : PageOrientation::Portrait)
{}


//////////////////////////////////////////////////////////////////////////
// OOMMF PostScript drawing class
const ClassDoc OommfPSDraw::class_doc("OommfPSDraw",
                    "Michael J. Donahue (michael.donahue@nist.gov)",
                    "1.0.1","12-Jun-2024");

OommfPSDraw::OommfPSDraw
(Tcl_Channel use_channel,
 OC_REAL8m pxoff,OC_REAL8m pyoff,
 OC_REAL8m printwidth,OC_REAL8m printheight,
 const char* pageorientation,
 OC_REAL8m xmin,OC_REAL8m ymin,
 OC_REAL8m xmax,OC_REAL8m ymax,
 OommfPackedRGB background_color,
 OC_REAL8m matwidth,
 OommfPackedRGB matcolor,
 OommfPackedRGB arrow_outline_color,
 OC_REAL8m arrow_outline_width,
 OC_REAL8m arrow_width,
 OC_REAL8m arrow_tip_width,
 OC_REAL8m arrow_length,
 OC_REAL8m arrow_head_width,
 OC_REAL8m arrow_head_ilen,
 OC_REAL8m arrow_head_olen)
: OommfDrawVectorGraphics
  (use_channel,pxoff,pyoff,printwidth,printheight,pageorientation,
   xmin,ymin,xmax,ymax,
   background_color,matwidth,matcolor,
   arrow_outline_color,arrow_outline_width,arrow_width,arrow_tip_width,
   arrow_length,arrow_head_width,arrow_head_ilen,arrow_head_olen)
{
  // Write header
  // NOTE: When using Nb_WriteChannel, each '%' in the output
  //  string is written to the output channel.  However, when
  //  using Nb_FprintfChannel, the passed string is treated as
  //  a format string, so to get a '%' on the output channel
  //  you need to use "%%".
  Nb_WriteChannel(channel,"%!PS-Adobe-3.0 EPSF-3.0\n",-1);
  Nb_WriteChannel(channel,"%%Creator: OommfPSDraw\n",-1);
  Nb_WriteChannel(channel,"%%CreationDate: ",-1);
  time_t current_time = time(NULL);
  Nb_WriteChannel(channel,ctime(&current_time),-1);
  Nb_FprintfChannel(channel,nullptr,1024,"%%%%BoundingBox: %d %d %d %d\n",
                    bbox_xmin,bbox_ymin,bbox_xmax,bbox_ymax);

  Nb_FprintfChannel(channel,nullptr,1024,"%%%%Orientation: %s\n",
                    pageorientation);

  Nb_WriteChannel(channel,"%%Pages: 1\n",-1);
  Nb_WriteChannel(channel,"%%DocumentData: Clean7Bit\n",-1);
  Nb_WriteChannel(channel,"%%EndComments\n\n",-1);

  // Base draw routines
  Nb_WriteChannel(channel,"%%BeginProlog\n",-1);

  // Position adjusment to device.  Intended for drawing lines
  Nb_WriteChannel(channel,"/DevHalfAdj { %stack: x y\n",-1);
  Nb_WriteChannel(channel,"  transform\n",-1);
  Nb_WriteChannel(channel,"  exch floor 0.5 add exch floor 0.5 add\n",-1);
  Nb_WriteChannel(channel,"  itransform\n",-1);
  Nb_WriteChannel(channel,"} def\n\n",-1);

  // Arrow routines. See NOTES IX, 27-Oct-2021, pp. 4-5 for details
  Nb_WriteChannel(channel,"/MakeStandardArrow { %stack: r g b\n",-1);
  if(ArrowOutlineWidth>0.0) { // Draw outline about standard arrow
    // Compute needed trig ratios
    // Special case handling: NOTES IX, 29-Oct-2021, pp. 6-7.
    // SPECIAL CASES: ArrowHeadWidth == ArrowTipWidth
    //                ArrowHeadOuterLength == 0
    //                ArrowHeadWidth == ArrowWidth
    // Convenience variables
    const OC_REAL8m Ld = ArrowHeadOuterLength - ArrowHeadInnerLength;
    const OC_REAL8m Lo = ArrowHeadOuterLength;
    const OC_REAL8m Wb = ArrowHeadWidth-ArrowWidth;    // width at back
    const OC_REAL8m Wt = ArrowHeadWidth-ArrowTipWidth; // width at tip
    const OC_REAL8m sAOW = ArrowOutlineWidth*ArrowWidth*0.25;

    // Compute cot(alpha2)+csc(alpha2) in NOTES IX
    const OC_REAL8m max_boff = sAOW + Ld;
    OC_REAL8m boff = sAOW*(2*Ld+sqrt(Wb*Wb+4*Ld*Ld));
    if(Ld<=0.0 || boff<Wb*max_boff) {
      boff /= Wb;
    } else {
      boff = max_boff;
    }


    // Special case handling at tip
    OC_REAL8m c2x=ArrowLength/2.-ArrowHeadOuterLength+sAOW;
    OC_REAL8m c2y=ArrowHeadWidth/2.+sAOW;
    OC_REAL8m dx =ArrowLength/2.+sAOW;
    OC_REAL8m dy =ArrowTipWidth/2.+sAOW;
    if(Wt>0.0 || Lo>0.0) {
      // Compute csc(alpha1)-cot(alpha1) in NOTES IX:
      OC_REAL8m coff = sAOW*Wt / (sqrt(Wt*Wt+4*Lo*Lo)+2*Lo);
      // Compute sec(alpha1)-tan(alpha1) in NOTES IX:
      OC_REAL8m doff = 2*sAOW*Lo / (sqrt(Wt*Wt+4*Lo*Lo)+Wt);
      c2x = ArrowLength/2.-ArrowHeadOuterLength+coff;
      dy =ArrowTipWidth/2.+doff;
    } else {
      dx = c2x = OC_MAX(dx,c2x);
      dy = c2y = OC_MAX(dy,c2y);
    }

    // Outline setup
    unsigned int red,green,blue;
    ArrowOutlineColor.Get(red,green,blue);
    Nb_FprintfChannel(channel,nullptr,1024,
                      "  %.6g %.6g %.6g setrgbcolor\n",
                      red/255.,green/255.,blue/255.);
    Nb_WriteChannel(channel,"  newpath\n",-1);

    // Render upper half of arrow outline
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g moveto\n",
         static_cast<double>(-ArrowLength/2.-sAOW),
         static_cast<double>(ArrowWidth/2.+sAOW));                  // Pt a
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(ArrowLength/2.-ArrowHeadInnerLength-boff),
         static_cast<double>(ArrowWidth/2.+sAOW));                  // Pt b
    if(Ld<=0.0 || Wb/2>0.75*sAOW) {
      Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(ArrowLength/2.-ArrowHeadOuterLength-boff),
         static_cast<double>(ArrowHeadWidth/2.+sAOW));              // Pt c1
      Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(c2x),static_cast<double>(c2y));        // Pt c2
    }
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(dx),static_cast<double>(dy));          // Pt d
    // Bottom half
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(dx),static_cast<double>(-dy));         // Pt d
    if(Ld<=0.0 || Wb/2>0.75*sAOW) {
      Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
         static_cast<double>(c2x),static_cast<double>(-c2y));       // Pt c2
      Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
        static_cast<double>(ArrowLength/2.-ArrowHeadOuterLength-boff),
        static_cast<double>(-(ArrowHeadWidth/2.+sAOW)));            // Pt c1
    }
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
        static_cast<double>(ArrowLength/2.-ArrowHeadInnerLength-boff),
        static_cast<double>(-(ArrowWidth/2.+sAOW)));                // Pt b
    Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
        static_cast<double>(-ArrowLength/2.-sAOW),
        static_cast<double>(-(ArrowWidth/2.+sAOW)));                // Pt a
    Nb_WriteChannel(channel,"  closepath\n  fill\n",-1);
  }
  // Render "standard" arrow (no translation, rotation, or scaling)
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);  // stack: r g b
  Nb_WriteChannel(channel,"  newpath\n",-1);
  // Top half
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g moveto\n",
                      static_cast<double>(-ArrowLength/2.),
                    static_cast<double>(ArrowWidth/2.));       // Pt A
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.-ArrowHeadInnerLength),
                      static_cast<double>(ArrowWidth/2.));     // Pt B
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.-ArrowHeadOuterLength),
                      static_cast<double>(ArrowHeadWidth/2.)); // Pt C
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.),
                      static_cast<double>(ArrowTipWidth/2.));  // Pt D
  // Do reflected bottom half
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.),
                      static_cast<double>(-ArrowTipWidth/2.));
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.-ArrowHeadOuterLength),
                      static_cast<double>(-ArrowHeadWidth/2.));
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(ArrowLength/2.-ArrowHeadInnerLength),
                      static_cast<double>(-ArrowWidth/2.));
  Nb_FprintfChannel(channel,nullptr,1024,"  %.6g %.6g lineto\n",
                      static_cast<double>(-ArrowLength/2.),
                      static_cast<double>(-ArrowWidth/2.));
  Nb_WriteChannel(channel,"  closepath\n  fill\n} def\n\n",-1);
  // MakeArrow wrapper routine that provides translation, rotation, and
  // scaling to MakeStandardArrow
  Nb_WriteChannel(channel,
                  "/MakeArrow { %stack: r g b xscale yscale rot xpos ypos\n",
                  -1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  translate\n  rotate\n  scale\n",-1);
  Nb_WriteChannel(channel,"  MakeStandardArrow\n",-1);
  Nb_WriteChannel(channel,"  grestore\n} def\n\n",-1);

  // Solid, filled rectangle
  Nb_WriteChannel(channel,"/SolidRect {"
                  " %stack: r g b x y width height\n",-1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  iydevscale mul /h exch def\n",-1);
  Nb_WriteChannel(channel,"  ixdevscale mul /w exch def\n",-1);
  Nb_WriteChannel(channel,"  moveto\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  xdevscale ydevscale scale\n",-1);
  Nb_WriteChannel(channel,"  currentpoint transform\n",-1);
  // Adjust rectangle base and side lengths to device pixels.
  Nb_WriteChannel(channel,"  exch dup w 0 gt\n",-1);
  Nb_WriteChannel(channel,
                  "  { floor sub dup w add ceiling /w exch def neg }\n",-1);
  Nb_WriteChannel(channel,
                  "  { ceiling sub dup w add floor /w exch def neg }\n",-1);
  Nb_WriteChannel(channel,"  ifelse\n",-1);
  Nb_WriteChannel(channel,"  exch dup h 0 gt\n",-1);
  Nb_WriteChannel(channel,
                  "  { floor sub dup h add ceiling /h exch def neg }\n",-1);
  Nb_WriteChannel(channel,
                  "  { ceiling sub dup h add floor /h exch def neg }\n",-1);
  Nb_WriteChannel(channel,"  ifelse\n",-1);
  Nb_WriteChannel(channel,"  rmoveto % Adjust corners to device pixels\n",-1);
  Nb_WriteChannel(channel,"  w 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  0 h rlineto\n",-1);
  Nb_WriteChannel(channel,"  w neg 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  fill\n",-1);
  Nb_WriteChannel(channel,"  grestore\n",-1);
  Nb_WriteChannel(channel,"} def\n\n",-1);


  // Routines for stipple fill
  Nb_WriteChannel(channel,"/StplRect {"
                  " %stack: r g b stipwidth stipheight"
                  " stipple x y width height\n",
                  -1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  4 2 roll moveto\n",-1);
  Nb_WriteChannel(channel,"  exch dup 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  exch 0 exch rlineto\n",-1);
  Nb_WriteChannel(channel,"  neg 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  clip\n",-1);
  Nb_WriteChannel(channel,"  /stipple exch def\n",-1);
  Nb_WriteChannel(channel,"  /stipheight exch def\n",-1);
  Nb_WriteChannel(channel,"  /stipwidth exch def\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  xdevscale ydevscale scale"
                  " % Adjust to device pixel size\n",-1);
  Nb_WriteChannel(channel,"  pathbbox %stack: x1 y1 x2 y2\n",-1);
  Nb_WriteChannel(channel,"  exch ceiling cvi\n",-1);
  Nb_WriteChannel(channel,"  exch ceiling cvi\n",-1);
  Nb_WriteChannel(channel,"  4 2 roll\n",-1);
  Nb_WriteChannel(channel,"  stipheight div floor cvi stipheight mul\n",-1);
  Nb_WriteChannel(channel,"  exch stipwidth div floor cvi stipwidth mul\n",-1);
  Nb_WriteChannel(channel,"  %stack: x2' y2' y1' x1'\n",-1);
  Nb_WriteChannel(channel,"  dup 2 index translate % Shift to stipple base\n",
                  -1);
  Nb_WriteChannel(channel,"  4 1 roll sub /h exch def\n",-1);
  Nb_WriteChannel(channel,"  sub neg 0 stipwidth 3 2 roll {\n",-1);
  Nb_WriteChannel(channel,"    pop\n",-1);
  Nb_WriteChannel(channel,"    stipwidth h true matrix {stipple} imagemask\n",
                  -1);
  Nb_WriteChannel(channel,"    stipwidth 0 translate\n",-1);
  Nb_WriteChannel(channel,"  } for\n",-1);
  Nb_WriteChannel(channel,"  grestore\n} def\n\n",-1);

#if PSDRAW_ALT_ARROWS
  // Dotted rectangle
  Nb_WriteChannel(channel,"/DotRect {"
                  " %stack: rI gI bI rO gO bO x y radius\n",-1);
  Nb_WriteChannel(channel,"  2 mul dup\n",-1);
  Nb_WriteChannel(channel,"  ixdevscale mul ceiling cvi /w exch def\n",-1);
  Nb_WriteChannel(channel,"  iydevscale mul ceiling cvi /h exch def\n",-1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  moveto\n",-1);
  Nb_WriteChannel(channel,"  xdevscale ydevscale scale"
                  " % Adjust to device pixel size\n",-1);
  Nb_WriteChannel(channel,"  w -2 div h -2 div rmoveto\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  currentpoint transform\n",-1);
  Nb_WriteChannel(channel,"  exch dup floor sub neg\n",-1);
  Nb_WriteChannel(channel,"  exch dup floor sub neg\n",-1);
  Nb_WriteChannel(channel,"  rmoveto % Move corner to device pixel\n",-1);
  Nb_WriteChannel(channel,"  currentpoint\n",-1);
  Nb_WriteChannel(channel,"  w 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  0 h rlineto\n",-1);
  Nb_WriteChannel(channel,"  w neg 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  fill\n",-1);
  Nb_WriteChannel(channel,"  moveto\n",-1);
  Nb_WriteChannel(channel,"  /wI w 3 div round cvi"
                  " dup w add 2 mod add def\n",-1);
  Nb_WriteChannel(channel,"  /hI h 3 div round cvi"
                  " dup w add 2 mod add def\n",-1);
  Nb_WriteChannel(channel,"  w wI sub 2 idiv h hI sub 2 idiv rmoveto\n",-1);
  Nb_WriteChannel(channel,"  wI 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  0 hI rlineto\n",-1);
  Nb_WriteChannel(channel,"  wI neg 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  fill\n",-1);
  Nb_WriteChannel(channel,"  grestore\n",-1);
  Nb_WriteChannel(channel,"} def\n\n",-1);
#endif // PSDRAW_ALT_ARROWS

#if !PSDRAW_ALT_ARROWS
  //  4 5 3 0 360 arc closepath

  // Dotted disk
  Nb_WriteChannel(channel,"/DotDisk {"
                  " %stack: rI gI bI rO gO bO x y radius\n",-1);
  Nb_WriteChannel(channel,"  dup\n",-1);
  Nb_WriteChannel(channel,"  /ro exch def\n",-1);
  Nb_WriteChannel(channel,"  4 div /ri exch def\n",-1);
  Nb_WriteChannel(channel,"  /cy exch def /cx exch def\n",-1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  cx cy ro 0 360 arc closepath\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor fill\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  cx cy ri 0 360 arc closepath\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor fill\n",-1);
  Nb_WriteChannel(channel,"  grestore\n",-1);
  Nb_WriteChannel(channel,"} def\n\n",-1);
#endif // !PSDRAW_ALT_ARROWS


  // Crossed diamond
#if 0 // Pre-202406 code. This confuses ps2pdf so that it produces very
      // skinny lines in the cross. Simplified version below seems OK.
  Nb_WriteChannel(channel,"/XDiamond {"
                  " %stack: rX gX bX rD gD bD x y radius\n",-1);
  Nb_WriteChannel(channel,"  dup\n",-1);
  Nb_WriteChannel(channel,"  ixdevscale mul ceiling cvi /rx exch def\n",-1);
  Nb_WriteChannel(channel,"  iydevscale mul ceiling cvi /ry exch def\n",-1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  moveto\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  xdevscale ydevscale scale"
                  " % Adjust to device pixel size\n",-1);
  Nb_WriteChannel(channel,"  currentpoint transform\n",-1);
  Nb_WriteChannel(channel,"  exch dup floor sub neg\n",-1);
  Nb_WriteChannel(channel,"  exch dup floor sub neg\n",-1);
  Nb_WriteChannel(channel,"  rmoveto % Move center to device pixel\n",-1);
  Nb_WriteChannel(channel,"  currentpoint\n",-1);
  Nb_WriteChannel(channel,"  rx neg 0 rmoveto\n",-1);
  Nb_WriteChannel(channel,"  rx ry neg rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx ry rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx neg ry rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  clip\n",-1);
  Nb_WriteChannel(channel,"  fill\n",-1);
  Nb_WriteChannel(channel,"  ry add moveto\n",-1);
  Nb_WriteChannel(channel,"  0 ry -2 mul rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx neg ry rmoveto\n",-1);
  Nb_WriteChannel(channel,"  rx 2 mul 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  rx ry add 14 div round cvi setlinewidth\n",-1);
  Nb_WriteChannel(channel,"  stroke\n",-1);
  Nb_WriteChannel(channel,"  grestore\n",-1);
  Nb_WriteChannel(channel,"} def\n",-1);
#endif // 0

  // Crossed diamond. Simplified version (202406) that ps2pdf groks
  Nb_WriteChannel(channel,"/XDiamond {"
                  " %stack: rX gX bX rD gD bD x y radius\n",-1);
  Nb_WriteChannel(channel,"  dup\n",-1);
  Nb_WriteChannel(channel,"  /ry exch def /rx exch def\n",-1);
  Nb_WriteChannel(channel,"  /cy exch def /cx exch def\n",-1);
  Nb_WriteChannel(channel,"  gsave\n",-1);
  Nb_WriteChannel(channel,"  newpath\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  cx cy moveto\n",-1);
  Nb_WriteChannel(channel,"  rx neg 0 rmoveto\n",-1);
  Nb_WriteChannel(channel,"  rx ry neg rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx ry rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx neg ry rlineto\n",-1);
  Nb_WriteChannel(channel,"  closepath\n",-1);
  Nb_WriteChannel(channel,"  clip\n",-1);
  Nb_WriteChannel(channel,"  fill\n",-1);
  Nb_WriteChannel(channel,"  setrgbcolor\n",-1);
  Nb_WriteChannel(channel,"  cx cy moveto\n",-1);
  Nb_WriteChannel(channel,"  rx neg 0 rmoveto\n",-1);
  Nb_WriteChannel(channel,"  rx 2 mul 0 rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx neg ry neg rmoveto\n",-1);
  Nb_WriteChannel(channel,"  0 ry 2 mul rlineto\n",-1);
  Nb_WriteChannel(channel,"  rx ry add 14 div setlinewidth\n",-1);
  Nb_WriteChannel(channel,"  stroke\n",-1);
  Nb_WriteChannel(channel,"  grestore\n",-1);
  Nb_WriteChannel(channel,"} def\n",-1);

  // Close header
  Nb_WriteChannel(channel,"%%EndProlog\n\n",-1);
  Nb_WriteChannel(channel,"%%BeginSetup\n",-1);
  Nb_WriteChannel(channel,"%%EndSetup\n\n",-1);

  Nb_WriteChannel(channel,"%%Page: 1 1\nsave\n\n",-1);

  unsigned int red,green,blue;
  background_color.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%% Background and clipping region\n"
                    "%.6g %.6g %.6g setrgbcolor\n",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%d %d moveto\n"
                    "%d %d lineto\n"
                    "%d %d lineto\n"
                    "%d %d lineto\n"
                    "closepath\nclip\nfill\n\n",
                    bbox_xmin,bbox_ymin,
                    bbox_xmax,bbox_ymin,
                    bbox_xmax,bbox_ymax,
                    bbox_xmin,bbox_ymax);

  if(page_orientation == PageOrientation::Landscape) {
    Nb_FprintfChannel(channel,nullptr,1024,
                      "%.6g %.6g translate\n",
                      static_cast<double>(bbox_xmax),
                      static_cast<double>(bbox_ymin));
    Nb_WriteChannel(channel,"90 rotate\n",-1);
  } else { // Portrait orientation
    Nb_FprintfChannel(channel,nullptr,1024,
                      "%.6g %.6g translate\n",
                      static_cast<double>(bbox_xmin),
                      static_cast<double>(bbox_ymin));
  }
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g scale\n",
                    static_cast<double>(scale),
                    static_cast<double>(-1*scale));
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g translate\n\n",
                    static_cast<double>(-xmin),
                    static_cast<double>(-ymax));

  Nb_WriteChannel(channel,"% Device pixel dimensions\n",-1);
  Nb_WriteChannel(channel,"1 1 idtransform /ydevscale exch def"
                  " /xdevscale exch def\n",-1);
  Nb_WriteChannel(channel,"/ixdevscale 1. xdevscale div def\n",-1);
  Nb_WriteChannel(channel,"/iydevscale 1. ydevscale div def\n\n",-1);
}

OommfPSDraw::~OommfPSDraw()
{ // Writes trailer
  Nb_WriteChannel(channel,"\ngrestore\n",-1);
  ApplyMat();
  Nb_WriteChannel(channel,"\nshowpage\n",-1);
  Nb_WriteChannel(channel,"%%Trailer\n%%EOF\n",-1);
}

void
OommfPSDraw::DrawPolyLine
(const Nb_List<PlanePoint>& vlist,
 OC_REAL8m linescale,
 OommfPackedRGB color,
 int joinstyle)
{ // Draws (potentially fat) line segments between each successive pair
  // of points in vlist.  If joinstyle == 0, then successive segments
  // are connected with a mitered join, == 1 gives a rounded joint, and
  // == 2 gives a beveled join.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those ends
  // will be joined using joinstyle.
  //
  // Note: The PostScript version of this routine includes code to
  //       adjust vertex locations to nearest whole (i.e., integral)
  //       point (/DevHalfAdj). Do we want that??

  const OC_INDEX pointcount = vlist.GetSize();
  if(pointcount<2) return; // Nothing to draw

  const PlanePoint *pta,*ptb;

  Nb_List_Index<PlanePoint> key;
  pta=vlist.GetFirst(key);

  Nb_FprintfChannel(channel,nullptr,1024,
                    "%d setlinejoin\n"
                    "newpath\n%.6g %.6g DevHalfAdj moveto\n",
                    joinstyle,
                    static_cast<double>(pta->x),
                    static_cast<double>(pta->y));

  OC_INDEX pointindex=1;
  while((ptb=vlist.GetNext(key))!=NULL) {
    ++pointindex;
    if(pointindex==pointcount && (*ptb)==(*pta)) {
      Nb_WriteChannel(channel,"closepath\n",-1);
    } else {
      Nb_FprintfChannel(channel,nullptr,1024,
                        "%.6g %.6g DevHalfAdj lineto\n",
                        static_cast<double>(ptb->x),
                        static_cast<double>(ptb->y));
    }
  }
  unsigned int red,green,blue;
  color.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g %.6g setrgbcolor\n",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g setlinewidth\n",
                    static_cast<double>(linescale));
  // Render line, and then reset join style to default (mitered).
  Nb_WriteChannel(channel,"stroke\n0 setlinejoin\n",-1);
}

void
OommfPSDraw::DrawFilledArrow
(OC_REAL8m xc,OC_REAL8m yc,
 OC_REAL8m size,
 OC_REAL8m xcos,OC_REAL8m ycos,OC_REAL8m zcos,
 OommfPackedRGB color)
{ // Draws arrow centered at point (xc,yc) in bitmap, of specified size,
  // with directional cosines xcos,ycos,zcos, in the specified color.
  unsigned int red,green,blue;
  color.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g %.6g %.6g %.6g"
                    " %.6g %.6g %.6g MakeArrow\n",
                    red/255.,green/255.,blue/255.,
                    static_cast<double>(size*sqrt(1.-zcos*zcos)),
                    static_cast<double>(size),
                    static_cast<double>(Oc_Atan2(ycos,xcos)*180./PI),
                    static_cast<double>(xc),
                    static_cast<double>(yc));
}

void
OommfPSDraw::DrawFilledRectangle
(OC_REAL8m x1,OC_REAL8m y1,
 OC_REAL8m x2,OC_REAL8m y2,
 OommfPackedRGB color,
 OC_REAL8m opaque)
{ // Opposing vertices (x1,y1) and (x2,y2) are included.  Opacity is
  // provided via stipple patterns. The opacity value should be in the
  // range [0,1], where 0 is fully transparent (no rectangle is drawn)
  // and 1 is fully opaque (i.e., no stipple). Other values are rounded
  // to the nearest of these available values:
  //   0.125, 0.25, 0.5, 0.75
  // Compare to the Tk gray12, gray25, gray50, and gray75 bitmasks.
#define MEMBERNAME "DrawFilledRectangle(4xOC_REAL8m,OommfPackedRGB,OC_REAL8m)"
  if(opaque<0.125/2) {
    return; // Empty fill
  }
  unsigned int red,green,blue;
  color.Get(red,green,blue);
  OC_REAL8m xmin = OC_MIN(x1,x2);
  OC_REAL8m xmax = OC_MAX(x1,x2);
  OC_REAL8m ymin = OC_MIN(y1,y2);
  OC_REAL8m ymax = OC_MAX(y1,y2);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  if(opaque<(0.125+0.25)/2) { // gray 12
    Nb_FprintfChannel(channel,nullptr,1024,
                      "8 4 <00110044> %.6g %.6g %.6g %.6g StplRect\n",
                      static_cast<double>(xmin),
                      static_cast<double>(ymin),
                      static_cast<double>(xmax-xmin),
                      static_cast<double>(ymax-ymin));
  } else if(opaque<(0.25+0.5)/2) { // gray25
    Nb_FprintfChannel(channel,nullptr,1024,
                      "8 2 <1144> %.6g %.6g %.6g %.6g StplRect\n",
                      static_cast<double>(xmin),
                      static_cast<double>(ymin),
                      static_cast<double>(xmax-xmin),
                      static_cast<double>(ymax-ymin));
  } else if(opaque<(0.5+0.75)/2) { // gray50
    Nb_FprintfChannel(channel,nullptr,1024,
                      "8 2 <55AA> %.6g %.6g %.6g %.6g StplRect\n",
                      static_cast<double>(xmin),
                      static_cast<double>(ymin),
                      static_cast<double>(xmax-xmin),
                      static_cast<double>(ymax-ymin));
  } else if(opaque<(0.75+1.0)/2) { // gray75
    Nb_FprintfChannel(channel,nullptr,1024,
                      "8 2 <77DD> %.6g %.6g %.6g %.6g StplRect\n",
                      static_cast<double>(xmin),
                      static_cast<double>(ymin),
                      static_cast<double>(xmax-xmin),
                      static_cast<double>(ymax-ymin));
  } else { // Solid fill
    Nb_FprintfChannel(channel,nullptr,1024,
                      "%.6g %.6g %.6g %.6g SolidRect\n",
                      static_cast<double>(xmin),
                      static_cast<double>(ymin),
                      static_cast<double>(xmax-xmin),
                      static_cast<double>(ymax-ymin));
  }
#undef MEMBERNAME
}

void
OommfPSDraw::DrawDiamondWithCross
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing out-of-plane vectors
#define MEMBERNAME "DrawDiamondWithCross(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  innercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  outercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g %.6g XDiamond\n",
                    static_cast<double>(xc),
                    static_cast<double>(yc),
                    static_cast<double>(InOutTipRadius*size*SQRT2));
#undef MEMBERNAME
}

#if PSDRAW_ALT_ARROWS
void
OommfPSDraw::DrawSquareWithDot
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing in-to-plane vectors.
#define MEMBERNAME "DrawSquareWithDot(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  innercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  outercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g %.6g DotRect\n",
                    static_cast<double>(xc),
                    static_cast<double>(yc),
                    static_cast<double>(InOutTipRadius*size));
#undef MEMBERNAME
}
#endif // PSDRAW_ALT_ARROWS

#if !PSDRAW_ALT_ARROWS
void
OommfPSDraw::DrawDiskWithDot
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing in-to-plane vectors.
#define MEMBERNAME "DrawDiskWithDot(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  innercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  outercolor.Get(red,green,blue);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g ",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,
                    "%.6g %.6g %.6g DotDisk\n",
                    static_cast<double>(xc),
                    static_cast<double>(yc),
                    static_cast<double>(InOutTipRadius*size*2/SQRTPI));
#undef MEMBERNAME
}
#endif // !PSDRAW_ALT_ARROWS




void
OommfPSDraw::ApplyMat()
{ // Lays flat mat around border of specified mat width
  // and color.
#define MEMBERNAME "ApplyMat(OC_REAL8m,OommfPackedRGB)"
  if(mat_width<=0.0) return;
  unsigned int red,green,blue;
  mat_color.Get(red,green,blue);
  Nb_WriteChannel(channel,"\n% Draw mat\n",-1);
  Nb_WriteChannel(channel,"gsave\nnewpath\n",-1);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g %.6g setrgbcolor\n",
                    red/255.,green/255.,blue/255.);
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g moveto\n",
                    static_cast<double>(bbox_xmin),
                    static_cast<double>(bbox_ymin));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmax),
                    static_cast<double>(bbox_ymin));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmax),
                    static_cast<double>(bbox_ymax));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmin),
                    static_cast<double>(bbox_ymax));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmin),
                    static_cast<double>(bbox_ymin));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmin+mat_width),
                    static_cast<double>(bbox_ymin));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmin+mat_width),
                    static_cast<double>(bbox_ymax-mat_width));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmax-mat_width),
                    static_cast<double>(bbox_ymax-mat_width));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmax-mat_width),
                    static_cast<double>(bbox_ymin+mat_width));
  Nb_FprintfChannel(channel,nullptr,1024,"%.6g %.6g lineto\n",
                    static_cast<double>(bbox_xmin),
                    static_cast<double>(bbox_ymin+mat_width));
  Nb_WriteChannel(channel,"closepath\nfill\ngrestore\n",-1);
#undef MEMBERNAME
}


//////////////////////////////////////////////////////////////////////////
// OOMMF SVG drawing class
const ClassDoc OommfSVGDraw::class_doc("OommfSVGDraw",
                    "Michael J. Donahue (michael.donahue@nist.gov)",
                    "1.0.1","12-Jun-2024");

OommfSVGDraw::OommfSVGDraw
(Tcl_Channel use_channel,
 OC_REAL8m pxoff,OC_REAL8m pyoff,
 OC_REAL8m printwidth,OC_REAL8m printheight,
 const char* pageorientation,
 OC_REAL8m xmin,OC_REAL8m ymin,
 OC_REAL8m xmax,OC_REAL8m ymax,
 OommfPackedRGB background_color,
 OC_REAL8m matwidth,
 OommfPackedRGB matcolor,
 OommfPackedRGB arrow_outline_color,
 OC_REAL8m arrow_outline_width,
 OC_REAL8m arrow_width,
 OC_REAL8m arrow_tip_width,
 OC_REAL8m arrow_length,
 OC_REAL8m arrow_head_width,
 OC_REAL8m arrow_head_ilen,
 OC_REAL8m arrow_head_olen)
: OommfDrawVectorGraphics
  (use_channel,pxoff,pyoff,printwidth,printheight,pageorientation,
   xmin,ymin,xmax,ymax,
   background_color,matwidth,matcolor,
   arrow_outline_color,arrow_outline_width,arrow_width,arrow_tip_width,
   arrow_length,arrow_head_width,arrow_head_ilen,arrow_head_olen)
{
  // Write header
  Nb_FprintfChannel(channel,nullptr,1024,
                    "<svg version=\"1.1\" xmlns=\"http://www.w3.org/2000/svg\"\n");
  int aw1 = bbox_xmin;
  int ah1 = bbox_ymin;
  int aw2 = bbox_xmax;
  int ah2 = bbox_ymax;
  if(page_orientation == PageOrientation::Landscape) {
    int t1 = aw1; aw1 = ah1; ah1 = t1;
    int t2 = aw2; aw2 = ah2; ah2 = t2;
  }
  Nb_FprintfChannel(channel,nullptr,1024,
                    " width=\"%d\" height=\"%d\" viewBox=\"%d %d %d %d\">\n",
                    aw1+aw2, ah1+ah2, 0, 0, aw1+aw2, ah1+ah2);

  // Background fill
  DrawFilledRectangle(aw1,ah1,aw2,ah2,background_color,1.0);

  // Clip box (user coordinates)
  Nb_FprintfChannel(channel,nullptr,1024,
       "<clipPath id=\"clipbox\">\n"
       " <rect x=\"%.8g\" y=\"%.8g\" width=\"%.8g\" height=\"%.8g\"/>\n"
       "</clipPath>\n",
       xmin,ymin,
       static_cast<double>((aw2-aw1)/scale),
       static_cast<double>((ah2-ah1)/scale));

  // Coordinate transform
  Nb_FprintfChannel(channel,nullptr,1024,
       "<g transform=\"translate(%d,%d) scale(%.8g) translate(%.8g,%.8g)\""
                    " clip-path=\"url(#clipbox)\">\n",
                    aw1,ah1,scale,-xmin,-ymin);
}

OommfSVGDraw::~OommfSVGDraw()
{
  // Write trailer
  Nb_WriteChannel(channel,"</g>\n",-1);
  ApplyMat();
  Nb_WriteChannel(channel,"</svg>\n",-1);
}

void
OommfSVGDraw::DrawPolyLine
(const Nb_List<PlanePoint>& vlist,
 OC_REAL8m linescale,
 OommfPackedRGB color,
 int joinstyle)
{ // Draws (potentially fat) line segments between each successive pair
  // of points in vlist.  If joinstyle == 0, then successive segments
  // are connected with a mitered join, == 1 gives a rounded joint, and
  // == 2 gives a beveled join.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those ends
  // will be joined using joinstyle.
  ColorMapNS::COLORSTR tkcolor;
  unsigned int red,green,blue;
  color.Get(red,green,blue);

  Nb_List_Index<PlanePoint> key;
  const PlanePoint *pta = nullptr;
  const PlanePoint *ptb = nullptr;

  const OC_INDEX pointcount = vlist.GetSize();
  if(pointcount<2) return; // Nothing to draw

  Nb_FprintfChannel(channel,nullptr,1024,
     "<path style=\"fill:none;stroke-width:%g;stroke-linecap:butt;"
     "stroke-linejoin:%s;stroke:#%02X%02X%02X;"
     "stroke-opacity:1;stroke-miterlimit:10;\"",
                    linescale,
                    (joinstyle==2 ? "bevel" :
                     (joinstyle==1 ? "round" :
                      "miter")), // Fallback default
                    red,green,blue);
  pta=vlist.GetFirst(key);
  Nb_FprintfChannel(channel,nullptr,1024," d=\"M %.8g %.8g",
                    static_cast<double>(pta->x),static_cast<double>(pta->y));
  OC_INDEX pointindex=1;
  while((ptb=vlist.GetNext(key))!=NULL) {
    ++pointindex;
    if(pointindex==pointcount && (*ptb)==(*pta)) {
      Nb_WriteChannel(channel," Z",-1); // Close path
    } else {
      Nb_FprintfChannel(channel,nullptr,1024," L %.8g %.8g",
                      static_cast<double>(ptb->x),static_cast<double>(ptb->y));
    }
  }
  Nb_WriteChannel(channel,"\" />\n",-1);
}

void
OommfSVGDraw::DrawFilledArrow
(OC_REAL8m xc,OC_REAL8m yc,
 OC_REAL8m size,
 OC_REAL8m xcos,OC_REAL8m ycos,OC_REAL8m zcos,
 OommfPackedRGB color)
{ // Draws arrow centered at point (xc,yc) in bitmap, of specified size,
  // with directional cosines xcos,ycos,zcos, in the specified color.
  // See NOTES IX 27-Oct-2021, pp. 4-7 and NOTES IX 272-Jun-2024,
  // pp. 70-71 for details.
  //
  // Note: An alternative to including the standard arrow vertex list in
  //  each file arrow instance would be to write it once and then
  //  reference it with a "use" command inside a group element with the
  //  appropriate transformation matrix.

  // Arrow tranformation (w/ <g> group)
  double a =  size*xcos;
  double b =  size*ycos;
  double c = 0.0;
  double d = 0.0;
  if(fabs(zcos) < 1.0) {
    double ixscrunch = sqrt(1.0-zcos*zcos);
    c = (-ycos/ixscrunch) * size;
    d =  (xcos/ixscrunch) * size;
  }
  std::stringstream ss;
  ss.precision(8);
  ss << "<g transform=\"matrix(" << a << " " << b << " " << c << " " << d
     << " " << xc << " " << yc << ")\">\n";

  if(ArrowOutlineWidth>0.0) {
    // Optional arrow outline requested. Fulfill request by drawing
    // a slighlty larger filled arrow shape, and the overwrite that
    // with the standard arrow. These are both drawn in the "standard
    // position" (i.e., aligned along +x), and adjusted via the matrix
    // transformation prescribed in the <g> element.

    // Compute needed trig ratios
    // Special case handling: NOTES IX, 29-Oct-2021, pp. 6-7.
    // SPECIAL CASES: ArrowHeadWidth == ArrowTipWidth
    //                ArrowHeadOuterLength == 0
    //                ArrowHeadWidth == ArrowWidth
    // Convenience variables
    const OC_REAL8m Ld = ArrowHeadOuterLength - ArrowHeadInnerLength;
    const OC_REAL8m Lo = ArrowHeadOuterLength;
    const OC_REAL8m Wb = ArrowHeadWidth-ArrowWidth;    // width at back
    const OC_REAL8m Wt = ArrowHeadWidth-ArrowTipWidth; // width at tip
    const OC_REAL8m sAOW = ArrowOutlineWidth*ArrowWidth*0.25;
    const double ax = -ArrowLength/2. - sAOW;
    const double ay = ArrowWidth/2. + sAOW;

    // Compute cot(alpha2)+csc(alpha2) in NOTES IX
    const OC_REAL8m max_boff = sAOW + Ld;
    OC_REAL8m boff = sAOW*(2*Ld+sqrt(Wb*Wb+4*Ld*Ld));
    if(Ld<=0.0 || boff<Wb*max_boff) {
      boff /= Wb;
    } else {
      boff = max_boff;
    }
    const double bx = ArrowLength/2.-ArrowHeadInnerLength-boff;
    const double by = ArrowWidth/2.+sAOW;

    // Special case handling at tip
    const double c1x = ArrowLength/2.-ArrowHeadOuterLength-boff;
    const double c1y = ArrowHeadWidth/2.+sAOW;

    double c2x=ArrowLength/2.-ArrowHeadOuterLength+sAOW;
    double c2y=ArrowHeadWidth/2.+sAOW;
    double dx =ArrowLength/2.+sAOW;
    double dy =ArrowTipWidth/2.+sAOW;
    if(Wt>0.0 || Lo>0.0) {
      // Compute csc(alpha1)-cot(alpha1) in NOTES IX:
      OC_REAL8m coff = sAOW*Wt / (sqrt(Wt*Wt+4*Lo*Lo)+2*Lo);
      // Compute sec(alpha1)-tan(alpha1) in NOTES IX:
      OC_REAL8m doff = 2*sAOW*Lo / (sqrt(Wt*Wt+4*Lo*Lo)+Wt);
      c2x = ArrowLength/2.-ArrowHeadOuterLength+coff;
      dy =ArrowTipWidth/2.+doff;
    } else {
      dx = c2x = OC_MAX(dx,c2x);
      dy = c2y = OC_MAX(dy,c2y);
    }

    // Vertex pts, upper half
    std::vector< std::pair<double,double> > pts {
      { ax, ay }, { bx, by }
    };
    if(Ld<=0.0 || Wb/2>0.75*sAOW) {
      pts.emplace_back(std::make_pair(c1x,c1y));  // Pt c1 (contingent)
      pts.emplace_back(std::make_pair(c2x,c2y));  // Pt c2 (contingent)
    }
    pts.emplace_back(std::make_pair(dx,dy));      // Pt d
    const int ptcount = static_cast<int>(pts.size());

    // Render outer arrow
    unsigned int outline_red,outline_green,outline_blue; // Outer color
    ArrowOutlineColor.Get(outline_red,outline_green,outline_blue);
    ss << " <path style=\"fill:#"
       << std::setfill('0') << std::setw(2) << std::hex << outline_red
       << std::setfill('0') << std::setw(2) << std::hex << outline_green
       << std::setfill('0') << std::setw(2) << std::hex << outline_blue
       << ";\"\n";
    ss << "  d=\"M "
       << pts[0].first << " " << pts[0].second; // Pt a
    for(int i=1; i<ptcount; ++i) { // Upper half
      ss << " L " << pts[i].first << " " << pts[i].second;
    }
    for(int i=ptcount-1; i>=0; --i) { // Lower half
      ss << " L " << pts[i].first << " " << -pts[i].second;
    }
    ss << " Z\"\n"; // "Z" ==> close standard arrow figure and fill
    ss << " />\n";
  }

  // Standard arrow vertex points, upper half
  std::vector< std::pair<double,double> > pts { // Upper half
    {-ArrowLength/2., ArrowWidth/2.},                           // Pt A
    { ArrowLength/2. -ArrowHeadInnerLength, ArrowWidth/2.},     // Pt B
    { ArrowLength/2. -ArrowHeadOuterLength, ArrowHeadWidth/2.}, // Pt C
    { ArrowLength/2., ArrowTipWidth/2.}                         // Pt D
  };
  const int ptcount = static_cast<int>(pts.size());

  // Render standard arrow
  // NB: In the " d=..." sequence, M is direct (non-relative) move,
  // D is direct (non-relative) draw, and Z closes and fills the
  // polygon region.
  unsigned int red,green,blue;  // Fill color
  color.Get(red,green,blue);
  ss << " <path style=\"fill:#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << ";\"\n";
  ss << "  d=\"M "
     << pts[0].first << " " << pts[0].second; // Pt A
  for(int i=1; i<ptcount; ++i) { // Upper half
    ss << " L " << pts[i].first << " " << pts[i].second;
  }
  for(int i=ptcount-1; i>=0; --i) { // Lower half
    ss << " L " << pts[i].first << " " << -pts[i].second;
  }
  ss << " Z\"\n"; // "Z" ==> close standard arrow figure and fill
  ss << " />\n";

  // Close <g> transformation group
  ss << "</g>\n";
  Nb_WriteChannel(channel,ss.str().c_str(),-1);
}

void
OommfSVGDraw::DrawFilledRectangle
(OC_REAL8m x1,OC_REAL8m y1,
 OC_REAL8m x2,OC_REAL8m y2,
 OommfPackedRGB color,
 OC_REAL8m opaque)
{ // Opposing vertices (x1,y1) and (x2,y2) are included.
#define MEMBERNAME "DrawFilledRectangle(4xOC_REAL8m,OommfPackedRGB,OC_REAL8m)"
  unsigned int red,green,blue;
  color.Get(red,green,blue);
  OC_REAL8m x = OC_MIN(x1,x2);
  OC_REAL8m width = OC_MAX(x1,x2) - x;;
  OC_REAL8m y = OC_MIN(y1,y2);
  OC_REAL8m height = OC_MAX(y1,y2) - y;;

  std::stringstream ss;
  ss.precision(8);
  ss << "<rect fill=\"#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << "\" x=\"" << x << "\" y=\"" << y
     << "\" width=\"" << width << "\" height=\"" << height
     << "\" opacity=\"" << opaque
     << "\"/>\n";

  Nb_WriteChannel(channel,ss.str().c_str(),-1);
#undef MEMBERNAME
}

void
OommfSVGDraw::DrawDiamondWithCross
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing out-of-plane vectors
#define MEMBERNAME "DrawDiamondWithCross(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  const double r = static_cast<double>(InOutTipRadius*size*SQRT2);

  std::stringstream ss;
  ss.precision(8);
  ss << "<g>\n";

  // Outer diamond
  outercolor.Get(red,green,blue);
  ss << " <path style=\"fill:#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << ";\"\n";
  ss << "  d=\"M " << xc << " " << yc;
  ss << " m " << -r << " 0"
     << " l " <<  r << " " << -r
     << " l " <<  r << " " <<  r
     << " l " << -r << " " <<  r;
  ss << " Z\" />\n";

  // Inside cross
  const double linewidth = r/7.0;
  const double cr = r - linewidth;
  innercolor.Get(red,green,blue);
  ss << " <path style=\"fill:none;stroke:#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << ";stroke-linecap:butt;stroke-width:" << linewidth
     << "\"\n";
  ss << "  d=\"M " << xc << " " << yc;
  ss << " m " <<  -cr << " 0"
     << " l " << 2*cr << " 0"
     << " m " <<  -cr << " " <<  cr
     << " l " <<   0  << " " << -2*cr;
  ss << " z\" />\n";

  ss << "</g>\n";
  Nb_WriteChannel(channel,ss.str().c_str(),-1);
#undef MEMBERNAME
}

#if PSDRAW_ALT_ARROWS
void
OommfSVGDraw::DrawSquareWithDot
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing in-to-plane vectors.
#define MEMBERNAME "DrawSquareWithDot(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  const double r = static_cast<double>(InOutTipRadius*size*SQRT2);
  const double ir = r/3.; // Inner radius

  std::stringstream ss;
  ss.precision(8);
  ss << "<g>\n";

  // Outer square
  outercolor.Get(red,green,blue);
  ss << " <path style=\"fill:#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << ";\"\n";
  ss << "  d=\"M " << xc << " " << yc;
  ss << " m " <<    -r << " " << -r
     << " l " <<   2*r << " 0"
     << " l " << " 0 " << 2*r
     << " l " <<  -2*r << " 0";
  ss << " Z\" />\n";

  // Inner square
  innercolor.Get(red,green,blue);
  ss << " <path style=\"fill:#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << ";\"\n";
  ss << "  d=\"M " << xc << " " << yc;
  ss << " m " <<    -ir << " " << -ir
     << " l " <<   2*ir << " 0"
     << " l " << " 0 " << 2*ir
     << " l " <<  -2*ir << " 0";
  ss << " Z\" />\n";

  ss << "</g>\n";
  Nb_WriteChannel(channel,ss.str().c_str(),-1);
#undef MEMBERNAME
}
#endif // PSDRAW_ALT_ARROWS

#if !PSDRAW_ALT_ARROWS
void
OommfSVGDraw::DrawDiskWithDot
(OC_REAL8m xc,OC_REAL8m yc,OC_REAL8m size,
 OommfPackedRGB outercolor,
 OommfPackedRGB innercolor)
{ // Symbol representing in-to-plane vectors.
#define MEMBERNAME "DrawDiskWithDot(3xOC_REAL8m,2xOommfPackedRGB)"
  unsigned int red,green,blue;
  const double r = static_cast<double>(InOutTipRadius*size*2/SQRTPI);
  // Set size to have same area as diamond with diagonal
  // InOutTipRadius*size*SQRT2

  const double ir = r/4.; // Inner radius

  std::stringstream ss;
  ss.precision(8);
  ss << "<g>\n";

  // Outer circle
  outercolor.Get(red,green,blue);
  ss << " <circle fill=\"#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << "\"";
  ss << " cx=\"" << xc << "\" cy=\"" << yc
     << "\" r=\"" << r << "\"/>\n";

  // Inner circle
  // Inner square
  innercolor.Get(red,green,blue);
  ss << " <circle fill=\"#"
     << std::setfill('0') << std::setw(2) << std::hex << red
     << std::setfill('0') << std::setw(2) << std::hex << green
     << std::setfill('0') << std::setw(2) << std::hex << blue
     << "\"";
  ss << " cx=\"" << xc << "\" cy=\"" << yc
     << "\" r=\"" << ir << "\"/>\n";

  ss << "</g>\n";
  Nb_WriteChannel(channel,ss.str().c_str(),-1);
#undef MEMBERNAME
}
#endif // !PSDRAW_ALT_ARROWS

void
OommfSVGDraw::ApplyMat()
{ // Lays flat mat around border of specified mat width
  // and color.
#define MEMBERNAME "ApplyMat(OC_REAL8m,OommfPackedRGB)"
  if(mat_width<=0.0) return;
  unsigned int red,green,blue;
  mat_color.Get(red,green,blue);

  double xmin = bbox_xmin;
  double ymin = bbox_ymin;
  double xmax = bbox_xmax;
  double ymax = bbox_ymax;
  double matw = mat_width;
  if(page_orientation == PageOrientation::Landscape) {
    double t1 = xmin; xmin = ymin; ymin = t1;
    double t2 = xmax; xmax = ymax; ymax = t2;
  }


  std::stringstream ss;
  ss.precision(8);

  ss << "<g fill=\"#"
     << std::hex
     << std::setfill('0') << std::setw(2) << red
     << std::setfill('0') << std::setw(2) << green
     << std::setfill('0') << std::setw(2) << blue
     << std::dec
     << "\">\n";

  // Bottom crosspiece
  ss << "<rect x=\"" << xmin << "\" y=\"" << ymin
     << "\" width=\""  << xmax - xmin
     << "\" height=\"" << matw << "\"/>\n";

  // Top crosspiece
  ss << "<rect x=\"" << xmin << "\" y=\"" << ymax - matw
     << "\" width=\""  << xmax - xmin
     << "\" height=\"" << matw << "\"/>\n";

  // Left crosspiece
  ss << "<rect x=\"" << xmin << "\" y=\"" << ymin
     << "\" width=\""  << matw
     << "\" height=\"" << ymax - ymin << "\"/>\n";

  // Right crosspiece
  ss << "<rect x=\"" << xmax - matw << "\" y=\"" << ymin
     << "\" width=\""  << matw
     << "\" height=\"" << ymax - ymin << "\"/>\n";

  ss << "</g>\n";
  Nb_WriteChannel(channel,ss.str().c_str(),-1);

#undef MEMBERNAME
}
