# File: oifconvert.tcl
# Purpose: Convert data block in OOMMF Integer Format (OIF) from one
#          representation (text, binary 1, binary 2, or binary 4) to
#          another.

set allow8byte 1  ;# 8-byte binary format not formally supported
                  ## as OIF format.

# Binary data block check values. These are designed to test
# byte order, accidental stripping of high bits, and channel
# sensitivity to control codes (1A = Ctrl-Z, 1C is ASCII file
# separator, and 04 is ASCII end of transmission). Also, FWIW,
# the CHECK4 and CHECK8 values are non-negative values in the
# corresponding signed integer type.
set CHECK1 0xFF
set CHECK2 0xFF1A
set CHECK4 0x04FF1A1C
set CHECK8 0x07060504FF1C1A08

proc Usage {{exitcode 0}} {
   if {$exitcode} {
      set chan stderr
   } else {
      set chan stdout
   }
   puts $chan "Usage: tclsh oifconvert.tcl <text|binary\
               <1|2|4>> <infile >outfile"
   exit $exitcode
}

switch -exact [lindex $argv 0] {
   text {
      if {[llength $argv]>1} {
         puts stderr "Error: Improper argument list"
         Usage 9
      }
      set outmode text
   }
   binary {
      if {[llength $argv] != 2} {
         puts stderr "Error: Improper argument list"
         Usage 9
      }
      set outmode binary
      set outwordwidth [lindex $argv 1]
      if {$outwordwidth != 1 && $outwordwidth != 2 && $outwordwidth != 4} {
         if {!$allow8byte || $outwordwidth != 8} {
            puts stderr "Error: Unsupported data word size: $outwordwidth"
            Usage 19
         }
      }
      set outwordwidth [expr {int(1*$outwordwidth)}] ;# Normalized form
   }
   {} -
   -h -
   --help { Usage }
   default {
      puts stderr "Error: Improper argument list"
      Usage 9
   }
}

set inchan stdin
set outchan stdout

set magic [gets $inchan]
if {![regexp {^#\s+OOMMF\s+OIF\s[0-9.]+} $magic]} {
   puts stderr "ERROR: Input file is not an OOMMF OIF file."
   exit 29
}

puts $outchan $magic
set checkline {}
set buffer {}
while {[gets $inchan line]>=0} {
   # Read and copy header up till data block header
   append buffer "$line\n"
   if {[string index $checkline end] == "\\"} {
      # Continue from last line; remove trailing backslace from checkline
      # and leading whitespace from next line.
      set checkline [string replace $checkline end end \
                        [regsub {^(#|)\s*} $line {}]]
   } else {
      set checkline $line
   }
   if {[string index $checkline end] != "\\"} {
      # Complete record
      set checkline [string tolower $checkline]
      if {[regexp {^#\s+begin\s*:\s+data\s+(.*)$} $checkline _ instyle]} {
         break
      }
      if {[regexp {^#\s+([xyz]nodes)\s*:\s*([0-9]+)} \
              $checkline _ component count]} {
         set ${component} $count
      }
      set checkline {}
      # The "Begin: data" record from the input file should not be
      # copied to outchan. To protect against a multi-line case like
      #   # Begin: \
      #   #   data text
      # we wait until we have a full record before outputting.
      puts -nonewline $outchan $buffer
      set buffer {}
   }
}

if {![info exists instyle]} {
   puts stderr "Error: No data block in input file"
   exit 39
}

if {![info exists xnodes] || ![info exists ynodes] || ![info exists znodes]} {
   puts stderr "Error: Incomplete or missing node count records"
   exit 49
}
if {$xnodes<1 || $ynodes<1 || $znodes<1} {
   puts stderr "Error: Invalid node count record"
   exit 49
}
set datasize [expr {$xnodes*$ynodes*$znodes}]

# Input data block
set inmode [lindex $instyle 0]
if {[string compare binary $inmode]==0} {
   set inwordwidth [lindex $instyle 1]
   if {$inwordwidth != 1 && $inwordwidth != 2 && $inwordwidth != 4} {
      if {!$allow8byte || $inwordwidth != 8} {
         puts stderr "Error: Unsupported input data block format: $inmode"
         exit 49
      }
   }
   set inwordwidth [expr {int(1*$inwordwidth)}] ;# Normalized form
} elseif {[string compare text $inmode]!=0} {
   puts stderr "Error: Unsupported input data block format: $instyle"
   exit 49
}
set data [list]

set trailer 0
if {[string compare text $inmode]==0} {
   # Text mode input
   while {[gets $inchan line]>=0} {
      set checkline [string tolower $line]
      if {[regexp {^#\s+end\s*:\s+data} $checkline]} {
         set trailer 1
         break
      }
      lappend data {*}$line
   }
   if {[llength $data] != $datasize} {
      puts stderr "Error: Data block size ([llength $data])\
                   doesn't match header ($datasize)"
      exit 59
   }
} else {
   # Binary input
   set saveconfig [fconfigure $inchan]
   fconfigure $inchan -translation binary
   switch -exact $inwordwidth {
      1 {
         set fmt "cu"  ;# One byte unsigned
         set endiancheck $CHECK1
      }
      2 {
         set fmt "su"  ;# Two byte unsigned little endian
         set endiancheck $CHECK2
      }
      4 {
         set fmt "iu"  ;# Four byte unsigned little endian
         set endiancheck $CHECK4
      }
      8 { ;# Not formally supported OIF format
         set fmt "wu"  ;# Eight byte unsigned little endian
         set endiancheck $CHECK8
      }
      default {
         error "Programming error: Unsupported word with $outwordwidth"
      }
   }
   # Test endian check value
   binary scan [read $inchan $inwordwidth] $fmt test
   if {$test != $endiancheck} {
      puts stderr "Error: Endian check failure on input file"
      exit 69
   }
   # Read values
   for {set i 0} {$i<$datasize} {incr i} {
      binary scan [read $inchan $inwordwidth] $fmt val
      lappend data $val
   }

   # Read trailer
   fconfigure $inchan {*}$saveconfig
   while {[gets $inchan line]>=0} {
      set checkline [string tolower [string trim $line]]
      if {[string length $checkline] == 0} { continue }
      if {[regexp {^#\s+end\s*:\s+data} $checkline]} {
         set trailer 1
         break
      }
      puts stderr "Error: Trailing characters in input data block"
      exit 79
   }
}
if {!$trailer} {
   puts stderr "WARNING: Input file missing data block trailer"
}

# Write out new data block
if {[string compare $outmode text]==0} {
   # Text mode
   puts $outchan "# Begin: data text"
   set dataindex 0
   foreach val $data {
      puts -nonewline $outchan " $val"
      if {[incr dataindex] % $xnodes == 0} {
         # Place newline at end of each xnodes output values.
         # Note that xnodes divides datasize, so this will
         # include a newline at the end of the block.
         puts $outchan {}
      }
   }
   puts $outchan "# End: data text"
} else {
   # Binary mode
   puts $outchan "# Begin: data binary $outwordwidth"
   set saveconfig [fconfigure $outchan]
   fconfigure $outchan -translation binary
   switch -exact $outwordwidth {
      1 {
         set fmt "c*"  ;# One byte
         set endiancheck $CHECK1
      }
      2 {
         set fmt "s*"  ;# Two byte little endian
         set endiancheck $CHECK2
      }
      4 {
         set fmt "i*"  ;# Four byte little endian
         set endiancheck $CHECK4
      }
      8 { ;# Not formally supported OIF format
         set fmt "w*"  ;# Eight byte little endian
         set endiancheck $CHECK8
      }
      default {
         error "Programming error: Unsupported word with $outwordwidth"
      }
   }
   puts -nonewline $outchan [binary format $fmt $endiancheck]
   puts -nonewline $outchan [binary format $fmt $data]
   fconfigure $outchan {*}$saveconfig
   puts $outchan "\n# End: data binary $outwordwidth"
}

# All done!
exit 0
