/* FILE: fileatlas.cc                 -*-Mode: c++-*-
 *
 * Atlas class derived from Oxs_Atlas class that uses an import
 * OIF (OOMMF Integer Format) file for region demarcation.
 */

#include "nb.h"
#include "director.h"
#include "fileatlas.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_FileAtlas);

/* End includes */

// Support routines
void Oxs_FileAtlas::SetupRegionMap
(const String& filename,
 Oxs_Box world)
{ // This routine should only be called AFTER processing
  // all the Specify block requests. This allows specifications
  // in the OIF to default fill missing info.

  // Read file
  int maxval = Vf_Oif::ReadFile(filename.c_str(),filedata);

  // If regions not set in Specify block, then use entries from file
  if(region_name.empty()) {
    // Region labels not included in Specify block. Use values
    // from file.
    if(filedata.labels.size()==0) {
      String errmsg = "Region labels not set in Specify block nor file \""
        + filename + "\"";
      throw Oxs_ExtError(errmsg);
    }
    region_name.push_back("universe"); // Implicit "0th" region
    int index = 0;
    if(filedata.labels[0].compare("universe")==0) {
      index = 1; // Skip (universe is included by default)
    }
    while(static_cast<size_t>(index)<filedata.labels.size()) {
      region_name.push_back(filedata.labels[index]);
      ++index;
    }
  }

  // Consistency check
  if(static_cast<size_t>(maxval) >= region_name.size()) {
    String errmsg = "Maximum data value ("
      + std::to_string(maxval)
      + " == " + std::to_string(maxval+1)
      + " values) overflows region list ("
      + std::to_string(region_name.size()) + " items)";
    throw Oxs_ExtError(this,errmsg);
  }

  // Region name -> id mapping
  // Throws an exception if a region name is associated with more than
  // one id. If we want to allow this then we could either rewrite
  // filedata.data after detecting a collision, or else switch
  // region_map to a multimap. In the latter case the GetRegionId(const
  // String& name) would need to be changed. Or, I guess we could
  // just ignore the collision and keep only the first mapping. (This
  // is effectively what imageatlas does.)
  region_map.clear();
  for(OC_INDEX i=0;static_cast<size_t>(i)<region_name.size();++i) {
    auto pair = region_map.emplace(region_name[i],i);
    if(!pair.second) {
      String errmsg = "Region " + region_name[i]
        + " assigned multiple id's";
      throw Oxs_ExtError(this,errmsg);
    }
  }

  // If world is empty, set world from file
  if(world.IsEmpty()) {
    if(!filedata.base_set || !filedata.stepsize_set) {
      String errmsg = "Atlas range not set in Specify block nor file \""
        + filename + "\"";
      throw Oxs_ExtError(errmsg);
    }
    world.Set(filedata.xbase - 0.5*filedata.xstepsize,
              filedata.xbase + filedata.xstepsize*(filedata.xnodes-0.5),
              filedata.ybase - 0.5*filedata.ystepsize,
              filedata.ybase + filedata.ystepsize*(filedata.ynodes-0.5),
              filedata.zbase - 0.5*filedata.zstepsize,
              filedata.zbase + filedata.zstepsize*(filedata.znodes-0.5));
    if(world.IsEmpty()) {
      String errmsg = "Atlas range from file \"" + filename + "\" is empty.";
      throw Oxs_ExtError(errmsg);
    }
  }

  // Set basept and scaling
  scale.Set(filedata.xnodes/(world.GetMaxX()-world.GetMinX()),
            filedata.ynodes/(world.GetMaxY()-world.GetMinY()),
            filedata.znodes/(world.GetMaxZ()-world.GetMinZ()));
  ThreeVector invscale(1.0/scale.x,1.0/scale.y,1.0/scale.z);
  basept.Set(world.GetMinX(),world.GetMinY(),world.GetMinZ());

  // Initialize region bounding boxes
  region_bbox.clear();
  region_bbox.push_back(world); // Set "universe" region to world

  // Initialize all non-universe region bboxes to empty boxes
  OC_INDEX id;
  Oxs_Box empty_box;
  for(id=1;id<static_cast<OC_INDEX>(region_name.size());id++) {
    region_bbox.push_back(empty_box);
  }

  // For each file node, expand corresponding region to
  // include the point in nodal coords.
  OC_INDEX index=0;
  for(OC_INDEX k=0;k<filedata.znodes;++k) {
    for(OC_INDEX j=0;j<filedata.ynodes;j++) {
      for(OC_INDEX i=0;i<filedata.znodes;i++) {
        Vf_Oif::DATA_TYPE tempid = filedata.data[index];
        if(tempid>0) { // Skip points in "universe" region.
          region_bbox[tempid].Expand(static_cast<OC_REAL8m>(i),
                                     static_cast<OC_REAL8m>(j),
                                     static_cast<OC_REAL8m>(k));
        }
        ++index;
      }
    }
  }

  // Convert bounding boxes from nodal to atlas coords.
  for(id=1;id<static_cast<OC_INDEX>(region_bbox.size());id++) {
    OC_REAL8m xmin = region_bbox[id].GetMinX();
    OC_REAL8m xmax = region_bbox[id].GetMaxX()+1.0; // Pad to full pixel
    OC_REAL8m ymin = region_bbox[id].GetMinY();
    OC_REAL8m ymax = region_bbox[id].GetMaxY()+1.0; // Pad to full pixel
    OC_REAL8m zmin = region_bbox[id].GetMinZ();
    OC_REAL8m zmax = region_bbox[id].GetMaxZ()+1.0; // Pad to full pixel
    xmin = xmin*invscale.x + basept.x;
    xmax = xmax*invscale.x + basept.x;
    ymin = ymin*invscale.y + basept.y;
    ymax = ymax*invscale.y + basept.y;
    zmin = zmin*invscale.z + basept.z;
    zmax = zmax*invscale.z + basept.z;
    region_bbox[id].Set(xmin,xmax,ymin,ymax,zmin,zmax);
  }
}

// Constructor
// NOTE: Scaling and translation options should be modeled after Oxs_FileVectorField
Oxs_FileAtlas::Oxs_FileAtlas
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr    // MIF block argument string
  ) : Oxs_Atlas(name,newdtr,argstr)
{
  // Process input string
  Oxs_Box world_box;
  if(HasInitValue("xrange")) {
    vector<OC_REAL8m> xrange; GetRealVectorInitValue("xrange",2,xrange);
    vector<OC_REAL8m> yrange; GetRealVectorInitValue("yrange",2,yrange);
    vector<OC_REAL8m> zrange; GetRealVectorInitValue("zrange",2,zrange);
    if(!world_box.CheckOrderSet(xrange[0],xrange[1],
                                yrange[0],yrange[1],
                                zrange[0],zrange[1])) {
      throw Oxs_ExtError(this,"One of [xyz]range in Specify block"
			 " is improperly ordered.");
    }
    if(world_box.GetVolume()<=0.0) {
      throw Oxs_ExtError(this,"Atlas bounding box has zero volume.");
    }
  }

  vector<String> regionlist;
  if(HasInitValue("regions")) {
      FindRequiredInitValue("regions",regionlist);
      if(regionlist.empty()) {
        throw Oxs_ExtError(this,"Empty region list");
      }
      // Copy regions list into region_name_list, setting special
      // "universe" region at id 0;
      region_name.push_back("universe");
      vector<String>::const_iterator it = regionlist.begin();
      if(it->compare("universe")==0) {
        // Already set, so skip
        ++it;
      }
      while(it != regionlist.end()) {
        if(it->compare("universe")==0) {
          throw Oxs_ExtError(this,
                             "universe is a reserved region label, and"
                             " must not be included in user region list.");
        }
        region_name.push_back(*it);
        ++it;
      }
      DeleteInitValue("regions");
  }

  String region_file = GetStringInitValue("file");

  VerifyAllInitArgsUsed();

  // Initialize region map
  SetupRegionMap(region_file,world_box);

  // Consistency checks
  if(region_bbox[0].GetVolume() <= 0.0) {
    throw Oxs_ExtError(this,"Atlas bounding box has zero volume.");
  }
  if(region_name.empty()) {
    throw Oxs_ExtError(this,"Empty region list");
  }
}


// Destructor
Oxs_FileAtlas::~Oxs_FileAtlas()
{}

OC_BOOL Oxs_FileAtlas::GetRegionExtents(OC_INDEX id,Oxs_Box &mybox) const
{ // If 0<id<GetRegionCount, then sets mybox to bounding box
  // for the specified region, and returns 1.  If id is 0,
  // sets mybox to atlas bounding box, and returns 1.
  // Otherwise, leaves mybox untouched and returns 0.
  if(id>=GetRegionCount()) return 0;
  mybox = region_bbox[id]; // Atlas bbox is stored in region_bbox[0].
  return 1;
}

OC_INDEX Oxs_FileAtlas::GetRegionId(const ThreeVector& point) const
{ // Returns the id number for the region containing point.
  // The return value is 0 if the point is not contained in
  // the atlas, i.e., belongs to the "universe" region.

  if(!region_bbox[0].IsIn(point)) return 0; // Outside atlas

  ThreeVector pt = point;
  pt -= basept;

  OC_INDEX i = static_cast<OC_INDEX>(floor(pt.x*scale.x));
  if(i>=filedata.xnodes) i = filedata.xnodes - 1;
  else if(i<0)           i = 0;

  OC_INDEX j = static_cast<OC_INDEX>(floor(pt.y*scale.y));
  if(j>=filedata.ynodes) j = filedata.ynodes - 1;
  else if(j<0)           j = 0;

  OC_INDEX k = static_cast<OC_INDEX>(floor(pt.z*scale.z));
  if(k>=filedata.znodes) k = filedata.znodes - 1;
  else if(k<0)           k = 0;

  OC_INDEX index = (k*filedata.ynodes + j) * filedata.xnodes + i;

  Vf_Oif::DATA_TYPE id = filedata.data[index];
  if(id>=static_cast<OC_UINDEX>(GetRegionCount())) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Programming error detected in "
		"Oxs_FileAtlas::GetRegionId(const ThreeVector&) const."
		"  Region id %lu discovered in file data array;"
		" max value should be %u",
                static_cast<unsigned long int>(id),
                static_cast<unsigned long int>(GetRegionCount()-1));
    throw Oxs_ExtError(this,buf);
  }

  return id;
}

OC_INDEX Oxs_FileAtlas::GetRegionId(const String& name) const
{ // Given a region id string (name), returns the corresponding region
  // id index.  If "name" is not included in the atlas, then -1 is
  // returned.  Note: If name == "universe", then the return value will
  // be 0.
  auto it = region_map.find(name);
  if(it != region_map.end()) {
    return it->second;
  }
  return -1;
}

OC_BOOL Oxs_FileAtlas::GetRegionName(OC_INDEX id,String& name) const
{ // Given an id number, fills in "name" with
  // the corresponding region id string.  Returns
  // 1 on success, 0 if id is invalid.  If id is 0,
  // then name is set to "universe", and the return
  // value is 1.
  name.erase();
  if(id>=GetRegionCount()) return 0;
  name = region_name[id];
  return 1;
}
