#!/bin/sh
# FILE: perfsupp.tcl               -*-Mode: tcl-*-
#
# Script to read output from a earlier perftest.tcl run
# and produce supplemental tables.
#
# NOTE: The derived tables may differ slightly from values
#       provided directly by perftest.tcl due to differences
#       in the precision of the TpFE values used in the
#       computation.
#
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}

#####################################################
### TODO: Handle input tables with missing values ###
#####################################################

proc Usage {} {
   puts "Augments perftest.tcl output with derived tables."
   puts "Usage: tclsh perfsupp.tcl <perftest.dat"
   exit 0
}

if {[llength $argv] != 0} {
   Usage
}

# Column headers
set cshead "Cell size (nm)"
set cchead "Cell count"
set pdhead "Pad ratio"
set colheads [list $cshead $cchead $pdhead]
set rphead "--- Threads ---" ;# Right panel upper header

proc CenterString { width string } {
   set len [string length $string]
   if {$len<$width} {
      set rw [expr {($width-$len)/2}]
      set lw [expr {$width-$len-$rw}]
      set string [format "%*s%s%*s" $lw {} $string $rw {}]
   }
   return $string
}

proc AddCommas { ival } {
   if {[regsub {(.)(...)$} $ival {\1,\2} ival]} {
      while {[regsub {([^,])(...,)} $ival {\1,\2} ival]} {}
   }
   return $ival
}

proc ValueWidths { thddata precision } {
   namespace import ::tcl::mathfunc::max
   set cellsizes  0
   set cellcounts 0
   set pads       0
   set times      0
   foreach elt $thddata {
      lassign $elt cs cc pd thdtimes
      lappend cellsizes $cs
      lappend cellcounts $cc
      lappend pads $pd
      lappend times {*}$thdtimes
   }

   set bigcellsize [max {*}$cellsizes]
   set cswidth     [string length $bigcellsize]

   set bigcellcount [max {*}$cellcounts]
   set ccwidth      [string length [AddCommas $bigcellcount]]

   set bigpad  [max {*}$pads]
   set pdwidth [string length [format %.2f $bigpad]]

   set bigtime    [max {*}$times]
   set timewidth  [string length [format "%.*f" $precision $bigtime]]

   return [list $cswidth $ccwidth $pdwidth $timewidth]
}

proc ColumnWidths { thdcnts valuewidths } {
   namespace import ::tcl::mathfunc::max
   lassign $valuewidths \
      cswidth ccwidth pdwidth timewidth

   set cswidth  [max $cswidth [string length $::cshead]]

   set ccwidth  [max $ccwidth [string length $::cchead]]

   set pdwidth [max $pdwidth [string length $::pdhead]]

   set bigthdcnts [max {*}$thdcnts]
   set timewidth  [max $timewidth [string length $bigthdcnts]]

   return [list $cswidth $ccwidth $pdwidth $timewidth]
}

proc DecimalPadCellsizes { thddata } {
   # For clarity of presentations, align values in the cell size
   # column at their (possibly implied) decimal point.
   namespace import ::tcl::mathfunc::max
   set maxdec 0
   foreach elt $thddata {
      set cs [string trim [lindex $elt 0]]
      if {[regexp {\.[0-9]*} $cs bits]} {
         set maxdec [max $maxdec [string length $bits]]
      }
   }
   # if any of the cellsizes have a decimal point, then
   # pad all cellsizes so string right alignment will
   # align decimal points
   if {$maxdec == 0} { return $thddata }
   foreach elt $thddata {
      set tail [lassign $elt cs]
      set cs [string trim $cs]
      if {![regexp {(|\.[0-9]*)$} $cs bits]} {
         error "Cell size format error: $cs"
      }
      append cs [string repeat " " [expr {$maxdec-[string length $bits]}]]
      lappend newthddata [linsert $tail 0 $cs]
   }
   return $newthddata
}

proc BaseTable {titlefmt tsttype thdcnts thddata precision} {
   # Generic formatting of thddata as a table
   set numthds [llength $thdcnts]
   if {$numthds<1} { error "No thread timing data" }

   # Special formatting for cell sizes.
   set thddata [DecimalPadCellsizes $thddata]

   # Calculate table width dimensions
   set valuewidths  [ValueWidths $thddata $precision]
   set lvwidths [lrange $valuewidths 0 2]
   set tvwidth [lindex $valuewidths 3]

   set columnwidths [ColumnWidths $thdcnts $valuewidths]
   set lwidths [lrange $columnwidths 0 2]
   set twidth [lindex $columnwidths 3]

   set rpheadstr " $::rphead "
   set rpwidth [expr {($twidth+2)*$numthds}] ;# Right panel width
   set overhang [expr {[string length $rpheadstr]-$rpwidth}]
   if {$overhang>0} {
      # Right panel header wider than natural data block
      set rpwidth [string length $rpheadstr]
      set colfiller [expr {$overhang/$numthds}]
      incr twidth $colfiller
      set extra [expr {$overhang - $numthds*$colfiller}]
      set rtmargin [expr {$extra/2}]
      set ltmargin [expr {$extra - $rtmargin}]
      set rtmargin [string repeat " " $rtmargin]
      set ltmargin [string repeat " " $ltmargin]
   } else {
      set rtmargin [set ltmargin {}]
   }

   set lpwidth 0  ;# Left panel width
   foreach w $lwidths t $::colheads {
      append colrow [format " %*s " $w [CenterString $w $t]]
      set lpwidth [expr {$lpwidth + $w + 2}]
   }
   # Pause creating colrow to compute edge and center margins

   # Determing margins and typeset title
   set tablewidth [expr {$lpwidth+$rpwidth+1}]
   set title [format $titlefmt $tsttype]
   set overhang [expr {[string length $title]-$tablewidth}]
   if {$overhang>0} {
      # Title is wider than natural table width. Split excess
      # across left, center bar, and right margins.
      set ctrmgn [expr {$overhang/4}]
      set lfullmgn [expr {($overhang-2*$ctrmgn)/2}]
      set rfullmgn [expr {$overhang-$lfullmgn-2*$ctrmgn}]

      set ldashmgn [string repeat "-" $lfullmgn]
      set lfullmgn [string repeat " " $lfullmgn]

      set rdashmgn [string repeat "-" $rfullmgn]
      set rfullmgn [string repeat " " $rfullmgn]

      set ctrbar [string repeat " " $ctrmgn]
      append ctrbar "|"
      append ctrbar [string repeat " " $ctrmgn]

      set ctrplus [string repeat "-" $ctrmgn]
      append ctrplus "+"
      append ctrplus [string repeat "-" $ctrmgn]

      incr tablewidth $overhang
   } else {
      set rfullmgn [set lfullmgn {}]
      set rdashmgn [set ldashmgn {}]
      set ctrbar "|"
      set ctrplus "+"
   }
   set titlerow "|[CenterString $tablewidth $title]|"

   # Finish creating colrow
   append colrow "$ctrbar$ltmargin"
   foreach t $::thdcnts {
      append colrow [format " %*s " $twidth [CenterString $twidth $t]]
   }
   append colrow "$rtmargin"
   set colrow "|$lfullmgn$colrow$rfullmgn|"


   set topseprow "+[string repeat - $tablewidth]+"
   set seprow "+$ldashmgn"
   append seprow \
      "[string repeat - $lpwidth]$ctrplus[string repeat - $rpwidth]"
   append seprow "$rdashmgn+"

   set uppercolrow "|$lfullmgn[string repeat " " $lpwidth]"
   append uppercolrow $ctrbar
   append uppercolrow "[CenterString $rpwidth $rpheadstr]"
   append uppercolrow "$rfullmgn|"

   set table "$topseprow\n$titlerow\n$seprow\n"
   append table "$uppercolrow\n$colrow\n$seprow\n"

   foreach rowdata $thddata {
      append table "|$lfullmgn"
      set thddata [lassign $rowdata cs cc pd]
      set cc [AddCommas $cc]
      # Left pane
      foreach cw $lwidths vw $lvwidths v [list $cs $cc $pd] {
         set v [format "%*s" $vw $v] ;# Right justify value
         append table [format " %*s " $cw [CenterString $cw $v]]
      }
      append table "$ctrbar$ltmargin"
      # Right (timings) pane
      foreach t {*}$thddata {
         set t [format "%*.*f" $tvwidth $precision $t] ;# Right justify value
         append table [format " %*s " $twidth [CenterString $twidth $t]]
      }
      append table "$rtmargin$rfullmgn|\n"
   }
   append table "$seprow"

   return $table
}

proc TpFE_Table { tsttype thdcnts thddata } {
   # Time per Field Evaluation table
   # No processing of thddata required
   set titlefmt " *** Test %s, time per field evaluation, in ms *** "
   set precision 1
   return [BaseTable $titlefmt $tsttype $thdcnts $thddata $precision]
}

proc TpFEpC_Table { tsttype thdcnts thddata } {
   # Time per Field Evaluation per Cell table
   # Efficiency table
   set newdata {}
   foreach rowdata $thddata {
      lassign $rowdata cs cc pd thdtimes
      set newvals {}
      foreach time $thdtimes {
         # Note: TpFE times are in milliseconds.
         set tpc [expr {1e6*$time/double($cc)}]
         lappend newvals $tpc
      }
      lappend newdata [list $cs $cc $pd $newvals]
   }
   set titlefmt " *** Test %s, time per field evaluation per cell, in ns *** "
   set precision 1
   return [BaseTable $titlefmt $tsttype $thdcnts $newdata $precision]
}

proc Efficiency_Table { tsttype thdcnts thddata } {
   # Efficiency table
   set newdata {}
   foreach rowdata $thddata {
      lassign $rowdata cs cc pd thdtimes
      # Efficiency in each row is scaled relative to the time 
      # and thread count for the first entry in the row
      set basethds [lindex $thdcnts 0]
      set basetime [lindex $thdtimes 0]
      set scaling [expr {double($basethds*$basetime)}]
      set newvals {}
      foreach thds $thdcnts time $thdtimes {
         set efc [expr {$scaling/($time*$thds)}]
         lappend newvals $efc
      }
      lappend newdata [list $cs $cc $pd $newvals]
   }
   set titlefmt " *** Test %s, threading efficiency *** "
   set precision 2
   return [BaseTable $titlefmt $tsttype $thdcnts $newdata $precision]
}

proc SpeedUp_Table { tsttype thdcnts thddata } {
   # Speed-up table
   set newdata {}
   foreach rowdata $thddata {
      lassign $rowdata cs cc pd thdtimes
      # Speed up in each row is scaled by the time and number of threads
      # relative to the first entry in the row
      set basethds [lindex $thdcnts 0]
      set basetime [lindex $thdtimes 0]
      set scaling [expr {double($basethds*$basetime)}]
      set newvals {}
      foreach time $thdtimes {
         set speedup [expr {$scaling/$time}]
         lappend newvals $speedup
      }
      lappend newdata [list $cs $cc $pd $newvals]
   }
   set titlefmt " *** Test %s, speed-ups *** "
   set precision 1
   return [BaseTable $titlefmt $tsttype $thdcnts $newdata $precision]
}

proc PrintTables {tsttype thdcnts thddata} {
   # puts [TpFE_Table $tsttype $thdcnts $thddata]
   puts {}
   puts [TpFEpC_Table $tsttype $thdcnts $thddata]
   puts {}
   puts [Efficiency_Table $tsttype $thdcnts $thddata]
   puts {}
   puts [SpeedUp_Table $tsttype $thdcnts $thddata]
   puts {}
}

set mode tableseek
set tsttype [set thdcnts [set thddata [list]]]
while {[gets stdin line]>=0} {
   puts $line
   switch -exact $mode {
      tableseek {
         if {[regexp \
              {[*]+ +Test *([^,]+), time per field evaluation, in ms +[*]+} \
                 $line _ tsttype]} {
            # Table found, and it's a "time per field evaluation" table
            set mode countseek
         }
      }
      countseek {
         # Search for thread count lines
         if {[regexp {\| *Cell size.*Pad ratio *\|([^|]*)\|} $line _ cnts]} {
            set thdcnts [list {*}$cnts]
            set mode dividerseek
         }
      }
      dividerseek {
         # Search for divider between header and result data
         if {[regexp {\+---*\+---*\+} $line]} {
            set  mode threaddata
         }
      }
      threaddata {
        if {[regexp {\| *([0-9.]+) +([0-9,]+) +([0-9.]+) *\| *([^|]*)\|} \
                 $line _ cellsize cellcount padratio threadtimes]} {
            # Record table data
            set cellcount [string map [list , {}] $cellcount]
            lappend thddata [list $cellsize $cellcount $padratio \
                                [list {*}$threadtimes]]
         } else {
            # Search for end of table data block
            if {[regexp {\+---[-+]*---\+} $line]} {
               # Table end found. Print results and search
               # for next table.
               set mode tableseek
               PrintTables $tsttype $thdcnts $thddata
               set thdcnts [set thddata [list]]
               set mode tableseek
            }
         }
         # Otherwise, ignore line?
      }         
   }
}
