# FILE: csourcefile.tcl
#
# The CSourceFile class -- instances of this class represent files of
# source code in the C/C++ language.  These instances are created to
# track the compile-time dependencies among them which imply what
# options are required to be passed to the compiler.

Oc_Class CSourceFile {

    # A map from the direct, absolute name of a file to the
    # CSourceFile instance which represents it.
    array common index

    common oommfRootDir
    common oommfExtDir

    const common configuration
    const common platformName

    # sysdepath is analogous to deppath, but for system directories.
    const common sysdeppath

    ClassConstructor {
        set configuration [Oc_Config RunPlatform]
        set platformName [$configuration GetValue platform_name]
        if {[catch {$configuration GetValue \
                        program_compiler_c++_system_include_path} \
                 sysdeppath]} {
            set sysdeppath {}
        }
        # Remove any trailing slashes from paths
        set tmppath [list]
        foreach _ $sysdeppath {
            lappend tmppath [string trimright $_ /]
        }
        set sysdeppath $tmppath

        set oommfRootDir [file normalize [Oc_Main GetOOMMFRootDir]]
        set oommfExtDir  [file join $oommfRootDir pkg]
    }

    # Direct pathname of this source code file
    const public variable name

    # List of directories to search for header files
    const public variable inc = {} {
        set tmp [list]
        foreach _ $inc {
            lappend tmp [Oc_DirectPathname $_]
        }
        set inc $tmp
    }

    # List of directories (direct paths) which contain files on which
    # this file depends (via recursive tracing of #include).
    private variable deppath

    # List of files that ultimately depend on this instance. Used to detect
    # circular dependencies.
    public variable depchain = {}
    public method ExtendDepChain { newchain } {
       foreach elt $newchain {
          if {[lsearch -exact $depchain $elt] == -1} {
             lappend depchain $elt
          }
       }
    }

    # Construction status. Used to detect circular dependencies.
    private variable done
    public method IsDone {} { return [info exists done] }

    # List of files (direct paths) on which this
    # file depends (via recursive tracing of #include)
    private variable dependencies

    Constructor {filename args} {
        set name [Oc_DirectPathname $filename]
        if {[info exists index($name)]} {
            set retval $index($name)
            $this Delete
            return $retval
        }
        set index($name) $this
        eval $this Configure $args
       $this FindIncludes ;# Note: FindIncludes includes
       ## $name on the dependencies list. Also, upon successful
       ## completion FindIncludes sets instance variable "done" true.
    }

    private proc PathMatch { elt pathlist } {
        # Uses the 'file nativename' call to see if
        # elt is in pathlist.  As compared to doing a
        # string-based comparison (a la lsearch), this
        # fixes some problems on Cygtel arising from
        # directory aliasing
        set elt [file nativename $elt]
        foreach name $pathlist {
            if {[string compare $elt [file nativename $name]]==0} {
                return 1
            }
        }
        return 0
    }

    private method AddDependency {newpath code} {
       if {[file readable $newpath]} {
          # Four cases:
          #   (0) Dependency conflict
          #       --> Throw circular dependency error
          #   (1) No existing CSourceFile instance for newpath
          #       --> Create a new one
          #   (2) CSourceFile instance exists and is fully constructed
          #       --> Return reference to existing instance
          #   (3) CSourceFile instance exists but is mid-contruction
          #       --> Call FindIncludes to finish construction
          #           (and cross fingers that recursive FindIncludes
          #           doesn't go BANG!)
          set newpath [Oc_DirectPathname $newpath]
          set wrkchain [linsert $depchain end $name]
          if {[lsearch -exact $wrkchain $newpath]>=0} {
             error "Circular dependency detected:\
                    \"$name\" ultimately #includes itself"
          }
          if {![info exists index($newpath)]} {
             # Create a new CSourceFile instance, with extended depchain
             $class New scf $newpath -inc $inc -depchain $wrkchain
          } else {
             set scf $index($newpath)
             $scf ExtendDepChain $wrkchain
             if {![$scf IsDone]} {
                # Putative CSourceFile instance is mid-construction.
                # Call FindIncludes and hope for the best
                $scf FindIncludes
             }
          }
          foreach npd [$scf Dependencies] {
             if {[lsearch -exact $dependencies $npd] == -1} {
                lappend dependencies $npd
             }
          }
          foreach npp [$scf DepPath] {
             if {![CSourceFile PathMatch $npp $deppath] && \
                    ![CSourceFile PathMatch $npp $sysdeppath]} {
                lappend deppath $npp
             }
          }
          return -code $code
       }
    }

    method FindIncludes {} {
       # This method determines all #include dependencies for file
       # $name. It does this by opening $name and scanning line-by-line
       # for '#include ...' statements. Each included file becomes a
       # dependency, and the included files themselves are recursively
       # checked to determine additional dependencies.
       #
       # So, first note that FindIncludes is re-entrant.
       #
       # Second, some system headers have #if/#elif/#endif macros that
       # select different #include files depending on local system
       # install conditions. The FindIncludes method makes no attempt to
       # parse preprocessor logic, instead if the #include'd file can be
       # found on the CSourceFile dependency search path it is blindly
       # thrown onto the dependency list. Otherwise it is dropped.
       #
       # A catch is that inside the OOMMF source tree some headers are
       # built by makerules.tcl files, so we want to handle missing
       # headers inside OOMMF specially.
       #
       # The code inside OOMMF is designed with pimake in mind, so there
       # is no thwarting preprocessor logic, and system headers are
       # always #include's using angle brackets, #include <filename.h>,
       # and non-system files with double quotes, #include "filename.h".
       # Third party code, however, doesn't necessary follow this rule.
       # For example, the sundials library used by the tuw_cvodeevolve
       # extension uses double quotes to #include some its files in
       # releases 6.x and 7.x. So to identify OOMMF local headers one
       # can check that the #include statement uses double quotes and is
       # inside a file in a directory under the OOMMF root directory.

       if {[catch {open $name r} f]} {
          error "Can't open C/C++ source code file '$name':\n   $f"
       }
       set dependencies [list $name]
       set deppath [linsert $inc 0 [file dirname $name]]

       # Remove system directories from deppath.
       set tmppath [list]
       foreach npp $deppath {
          if {![CSourceFile PathMatch $npp $sysdeppath]} {
             lappend tmppath $npp
          }
       }
       set deppath $tmppath

       # Set local_parent to 1 iff working file "name" is under
       # the OOMMF directory hierarchy
       set local_parent 0
       if {[string equal -length [string length $oommfRootDir] \
               $oommfRootDir [file normalize $name]]} {
          set local_parent 1
       }

       set code [catch {
          while {[gets $f line] >= 0} {
             # For each line in $f:
             #   (1) If matches "End includes", break.
             #   (2) If doesn't match '#include' syntax, continue.
             #   (3) Determine if file is a "local" (i.e., OOMMF)
             #       or "non-local" (i.e., system) include file.
             #   (4) If "local" include then search oommf directory tree
             #       and makerules. If match found, record and continue.
             #       Otherwise, error.
             #   (5) System #includes
             #     (5a) If file is {tcl,tk}.h, record and continue.
             #     (5b) Else go to (6).
             #   (6) Search "extra" include paths; record if found.
             #   (7) Continue.
             #
             # Note 1: In the above, "break" means break out of
             #    while-gets-line loop, and continue means jump to next
             #    gets line.
             #
             # Note 2: In principle step 3b should shoule always throw
             #    an error, because all local includes should be
             #    traceable.  Unfortunately, third party code is outside
             #    of our control and some times bad code happens (i.e.,
             #    system files may be #include'd with double quotes
             #    instead of angle brackets).

             # STEP (1) If matches "End includes", break
             if {[regexp -nocase \
                     "/\\\*\[ \t\]*end\[ \t\]+includes?\[ \t\]*\\\*/" \
                     $line]} {
                # The above is a liberal match against lines including
                # the sequence '/* End includes */'
                break
             }

             # STEP (2) If doesn't match '#include' syntax, continue
             if {![regexp "^\[ \t\]*#\[ \t\]*include\[ \t\]+(<)(\[^>\]+)>" \
                      $line match include_type newfile] &&
                 ![regexp "^\[ \t\]*#\[ \t\]*include\[ \t\]+(\")(\[^>\]+)\"" \
                      $line match include_type newfile]} {
                # Not an '#include' line
                continue
             }
             set newfile [string trim $newfile]

             # STEP (3) Distinguish local (oommf) from non-local
             #          (system) include files
             set local_include 0
             if {[string compare \" $include_type]==0 && $local_parent} {
                # If newfile is taken relative to parent file directory,
                # would it lie inside the OOMMF file hierarchy?
                set testname [file join [file dirname $name] $newfile]
                if {[string equal -length [string length $oommfRootDir] \
                        $oommfRootDir $testname]} {
                   set local_include 1
                }
             }

             # STEP (4) Local includes. Search oommf directory tree and
             #        makerules. If match found, record and continue.
             #        Otherwise throw and error.
             if {$local_include} {
                set searchpath {}
                foreach dir $deppath {
                   lappend searchpath $dir
                   set newpath [file join $dir $newfile]
                   $this AddDependency $newpath break
                   lappend searchpath [file join $dir $platformName]
                   set newpath [file join $dir $platformName $newfile]
                   $this AddDependency $newpath break
                }
                if {[info exists index($newpath)]} { continue }

                # Haven't yet found the #included file.
                if {[regexp {^([^.]*)\.h$} [file tail $newfile] _ stem]} {
                   # Assume it is the public header file of an extension.
                   # By convention, it has the same name as that
                   # extension.  For now, the directory containing that
                   # extension also has that name.  Later we may need to
                   # deal with version numbers and resolving the most
                   # recent extension release at this point.
                   set dir [file join $oommfExtDir $stem]
                   if {[file isdirectory $dir]} {
                      lappend searchpath $dir
                      set newpath [file join $dir $newfile]
                      $this AddDependency $newpath ok
                   }
                }
                if {[info exists index($newpath)]} { continue }

                # Couldn't find the #include file.
                # See if there's a MakeRule target to build it.
                # Search in same places we searched for files.
                foreach dir $deppath {
                   set newpath [file join $dir $newfile]
                   if {[string length [MakeRule FindRule $newpath]]} {
                      if {[lsearch -exact $dependencies $newpath] == -1} {
                         lappend dependencies $newpath
                      }
                      break
                   }
                   set newpath [file join $dir $platformName $newfile]
                   if {[string length [MakeRule FindRule $newpath]]} {
                      if {[lsearch -exact $dependencies $newpath] == -1} {
                         lappend dependencies $newpath
                      }
                      set npp [file join $dir $platformName]
                      if {[lsearch -exact $deppath $npp] == -1 && \
                             [lsearch -exact $sysdeppath $npp] == -1} {
                         lappend deppath $npp
                      }
                      break
                   }
                }
                if {[string length [MakeRule FindRule $newpath]]} { continue }
                # Included file not found.
                error "Can't find #include dependency \"$newfile\" located\
                       in file '$name'.\nSearched the following\
                       directories: $searchpath"
             }

             # STEP (5) System includes
             #       (5a) If file is {tcl,tk}.h, record and continue
             #       (5b) Else go to (6)
             if {[string compare tcl.h $newfile]==0 \
                    || [string compare tk.h $newfile]==0} {
                if {[catch {$configuration GetValue use_tk} use_tk]} {
                   set use_tk 1   ;# Default is to build with Tk
                }
                if {[string match tcl.h $newfile]} {
                   global env
                   set includeDir $env(OOMMF_TCL_INCLUDE_DIR)
                } elseif {$use_tk && [string match tk.h $newfile]} {
                   global env
                   set includeDir $env(OOMMF_TK_INCLUDE_DIR)
                }
                if {$use_tk || ![string match tk.h $newfile]} {
                   # Don't include tk.h in dependencies if use_tk is false
                   set newpath [file join $includeDir $newfile]
                   if {[lsearch -exact $dependencies $newpath] == -1} {
                      lappend dependencies $newpath
                   }
                   if {[lsearch -exact $deppath $includeDir] == -1 && \
                          [lsearch -exact $sysdeppath $includeDir] == -1} {
                      # Be aggressive about putting Tcl and Tk directories
                      # at front of include list
                      set deppath [linsert $deppath 0 $includeDir]
                   }
                }
                if {$use_tk && [string match tk.h $newfile]} {
                   set xinc [$configuration GetValue TK_XINCLUDES]
                   if {[string match "\#*" $xinc]} {
                      # File tkConfig.sh in versions of Tk prior to 8.4 have
                      #    TK_XINCLUDES='# no special path needed'
                      set xinc {}
                   }
                   foreach npp $xinc {
                      regsub -- {^-I} $npp {} npp
                      set npp [string trim $npp \"]
                      if {[lsearch -exact $deppath $npp] == -1 \
                             && [lsearch -exact $sysdeppath $npp] == -1 \
                             && ![string match #* $npp] \
                             && [string length $npp]} {
                         lappend deppath $npp
                      }
                   }
                }
                continue  ;# Read next line in file
             }

             # STEP (6) Search "extra" include paths; record if found. We add
             #       these only as a last resort, and at the end, to
             #       minimize risk of interference with system include
             #       directories (in particular, we don't want to look for
             #       Tcl/Tk headers in these directories).
             if {![catch {$configuration GetValue \
                             program_compiler_extra_include_dirs} idirs]} {
                set newpath {}
                foreach npp $idirs {
                   set npp [Oc_DirectPathname $npp]
                   regsub -- {^-I} $npp {} npp
                   set npp [string trim $npp \"]
                   if {[lsearch -exact $deppath $npp] == -1 \
                          && [lsearch -exact $sysdeppath $npp] == -1 \
                          && ![string match #* $npp] \
                          && [string length $npp]} {
                      set newpath [file join $npp $newfile]
                      $this AddDependency $newpath break
                      set newpath [file join $npp $platformName $newfile]
                      $this AddDependency $newpath break
                   }
                }
                if {[string length $newpath] && \
                       [info exists index($newpath)]} {
                   # Found it
                   lappend deppath $npp
                   continue
                }
                # Ignore missing "system" includes.
             }
             # STEP (7): Not found; continue back to top of while loop
          } ;# End 'while {[gets $f line] >= 0} { ... }'
       } msg] ;# End 'set code [catch { ... } ]
       catch {close $f}
       if {$code} {
          error "Error finding includes in $name:\n  $msg"
       }
       set done 1
    }

    Destructor {
        if {[info exists name]} {
            if {[string match $this $index($name)]} {
                unset index($name)
            }
        }
    }

    method Dependencies {} {
        if {[info exists done]} {
            return $dependencies
        }
        error "Circular dependency detected:
'$name' ultimately #includes itself"
    }

    method DepPath {} {
        return $deppath
    }

}
