# FILE: procs.tcl
#
# A collection of Tcl procedures used by pimake application

set oommfExtDir [file join [file dirname [file dirname [file dirname \
                [Oc_DirectPathname [info script]]]]] pkg]

proc DeleteFiles {args} {
    if {![llength $args]} {return}
    puts "Deleting $args in [pwd] ..."
    file delete -force -- {*}$args
}

proc MakeDirectory {dir} {
    puts "Making directory '[file join [pwd] $dir]' ..."
    file mkdir $dir
}

proc RenameFile { source target } {
   puts "Renaming $source to $target ..."
   file rename -force -- $source $target
}

proc VerifyExecutable {stemlist {noexedelete 0}} {
    foreach fn [Platform Executables $stemlist] {
        if {![file executable $fn]} {
           if {$noexedelete} {
              puts "WARNING: $fn is not executable."
           } else {
              puts "$fn not executable.  Deleting ..."
              file delete $fn
           }
        } else {
        }
    }
}

interp alias {} MakeTclIndex {} Oc_MakeTclIndex

# Create automatic dependencies one level below the current directory.
# This proc should only be used with default targets (e.g., "all",
# "configure", "clean", ...) found in makerules.config as there is no
# checking here that the constructed targets actually exist. The
# straightforward method to check for target existence would be to load
# the subdirectory makerules.tcl, but that would likely be a recursive
# call into the makerules.tcl loading code, since RecursiveDeps itself
# is called from the makerules.tcl load process in the current
# directory. This may re-order the MakeRule object instance creation
# order and muck about with MakeRule class-level initialization.
proc RecursiveDeps { target } {
   set deps [list]
   foreach mr [glob -nocomplain [file join * makerules.tcl]] {
      lappend deps [file join [file dirname [Oc_DirectPathname $mr]] $target]
   }
   return $deps
}

# Routine to create Oxs_Exs extension initialization (*.cc) file
# The import outfile is the name of the output file; args is a
# list of files to scan for OXS_EXT_CHILD_INTERFACE_SUPPORT macros.
proc MakeOxsExtInit { outfile args } {
    puts "Creating [file join [pwd] $outfile] ..."
    set cnamelist {}
    # Scan files
    foreach file $args {
	if {[catch {open $file "r"} chan]} {
	    continue              ;# Bad filename; skip
	}
	# Read file up to '/* End includes */' or eof.
	set buf {}
	while {[gets $chan line]>=0} {
	    if {[regexp -nocase \
                    "/\\\*\[ \t\]*end\[ \t\]+includes?\[ \t\]*\\\*/" \
                    $line]} {
		# Line matched '/* End includes */'
		break
	    }
	    append buf $line
	    append buf "\n"
	}
	catch {close $chan}

	# Remove backslash+newline combinations
	regsub -all -- "\\\\\n" $buf {} buf

	# Search for OECIS macro
	set ws "\[ \t\n\]*"
	set bs "\[^ \t\b\]*"
	set key "OXS_EXT_CHILD_INTERFACE_SUPPORT"
	if {[regexp \
		"$key$ws\\($ws\($bs\)$ws\\)" \
		$buf dummy cname]} {
	    lappend cnamelist $cname
	}
    }

    # Write out extension initialization file
    if {![catch {open $outfile "w"} outchan]} {
        set declfmt {void OxsExtGet%s}
        append declfmt {RegistrationInfo(Oxs_ExtRegistrationBlock&);}
        set callfmt {  OxsExtGet%sRegistrationInfo(regblk);}
        puts $outchan "\#include \"ext.h\""
        puts $outchan "\n/* End includes */\n"
        foreach cname $cnamelist {
            puts $outchan [format $declfmt $cname]
        }
        puts $outchan "\nvoid OxsInitializeExt()\n{"
        puts $outchan {  Oxs_ExtRegistrationBlock regblk;}
        foreach cname $cnamelist {
            puts $outchan [format $callfmt $cname]
            puts $outchan {  Oxs_Ext::Register(regblk);}
        }
        puts $outchan "}"
        catch {close $outchan}
    }
}

# Restricted* procs: These are intended to give read-only access to the
# system file system, to be used by Oxs extension .rules files.  These
# .rules files are processed inside child interpreters by
# oxs/makerules.tcl. The intention of these procs is to protect against
# accidental corruption of the build process by errors in the .rules
# scripts. These procs don't provide any system access not already
# available to the Oxs extension C++ code, so the .rules files should be
# reviewed in a like manner.
#
#    In particular, the Restricted* procs are NOT designed for use
#    inside .mif scripts.
#
proc RestrictedFile {cmd args} {
   # Restricted version of Tcl 'file' command that allows some read-only
   # accesses.
   set allowed_commands {
      exists
      isdirectory
      join
   }
   if {[lsearch -exact $allowed_commands $cmd]<0} {
      error "RestrictedFile \"$cmd\" not supported"
   }
   return [file $cmd {*}$args]
}

proc RestrictedDirGlob { dir globstr } {
   # Version of Tcl 'glob' restricted to directory "$dir"
   set matchlist {}
   set filelist [glob -nocomplain -directory $dir *]
   foreach file $filelist {
      if {[string match $globstr $file]} {
         lappend matchlist $file
      }
    }
    return $matchlist
}

proc RestrictedFileGrep { filename matchcount args} {
   set matches [list]
   set chan [open $filename r]
   while {[gets $chan line]>=0} {
      foreach pat $args {
         set match [regexp -inline -- $pat $line]
         if {[llength $match]>0} {
            lappend matches $match
            if {$matchcount && [llength $matches]>=$matchcount} {
               break
            }
         }
      }
      if {$matchcount && [llength $matches]>=$matchcount} {
         break
      }
   }
   close $chan
   return $matches
}
