/* FILE: colorquantity.cc                 -*-Mode: c++-*-
 *
 * Color quantities for use by Vf_Mesh objects.
 *
 */

#include <cmath>

#include "colorquantity.h"

/* End includes */     /* Optional directive to build.tcl */

std::unordered_map
 <std::string,Vf_ColorQuantity::ColorQuantityAttributes>
  Vf_ColorQuantity::registered_quantities;

std::list<std::string> Vf_ColorQuantity::registered_names;

namespace {  // Unnamed namespace
  // Note: Don't include "none" on the base quantity list.  Instead,
  // leave the mesh responsible for "none"; that way, the mesh can place
  // "none" at the end of the quantity list, and its GetDisplayList
  // routine can shortcircuit quantity evaluation.

  void GetVecShade_x
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  { shade[0] = static_cast<OC_REAL4>((1.+v.x)/2); }
  Vf_ColorQuantity GVS_x("x",GetVecShade_x);
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_y
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  { shade[0] = static_cast<OC_REAL4>((1.+v.y)/2); }
  Vf_ColorQuantity GVS_y("y",GetVecShade_y);
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_z
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  { shade[0] = static_cast<OC_REAL4>((1.+v.z)/2); }
  Vf_ColorQuantity GVS_z("z",GetVecShade_z);
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_mag
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  { shade[0] = static_cast<OC_REAL4>(sqrt(v.x*v.x + v.y*v.y + v.z*v.z)); }
  Vf_ColorQuantity GVS_mag("mag",GetVecShade_mag);
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_xyangle
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  {
    if(v.x==0. && v.y==0.) {
      shade[0] = 0.0;
    } else {
      shade[0] = static_cast<OC_REAL4>(Oc_Atan2(v.y,v.x)*(1./(2*PI)));
      /// Range [-0.5:0.5]
      shade[0] = shade[0] - floor(shade[0]);    // Force into [0,1)
    }
    OC_REAL4 z = static_cast<OC_REAL4>((v.z+1)/2);
    shade[1] = (z<0 ? 0 : (z>1 ? 1 : z)); // Force into [0,1]
#if 0 // DEBUG
    std::cerr << "XYANG: (" << v.x << "," << v.y << ","<< v.z
              << ") -> [" << shade[0] << "," << shade[1] << "]"
              << std::endl;
#endif// 0 DEBUG
  }

  Vf_ColorQuantity GVS_xyangle("xy-angle",
                               {GetVecShade_xyangle,
                                Vf_ColorQuantity::EdgeType::Periodic});
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_xzangle
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  {
    if(v.z==0. && v.x==0.) {
      shade[0] = 0.0;
    } else {
      shade[0] = static_cast<OC_REAL4>(Oc_Atan2(v.z,v.x)*(1./(2*PI)));
      /// Range [-0.5:0.5]
      shade[0] = shade[0] - floor(shade[0]);    // Force into [0,1)
    }
    OC_REAL4 y = static_cast<OC_REAL4>((v.y+1)/2);
    shade[1] = (y<0 ? 0 : (y>1 ? 1 : y)); // Force into [0,1]
  }
  Vf_ColorQuantity GVS_xzangle("xz-angle",
                               {GetVecShade_xzangle,
                                Vf_ColorQuantity::EdgeType::Periodic});
  /// Instantiation registers name + {function, edgetype}

  void GetVecShade_yzangle
  (const Nb_Vec3<OC_REAL4>& v,
   Vf_DisplayVector::SHADEVEC_T& shade)
  {
    if(v.z==0. && v.y==0.) {
      shade[0] = 0.0;
    } else {
      shade[0] = static_cast<OC_REAL4>(Oc_Atan2(v.z,v.y)*(1./(2*PI)));
      /// Range [-0.5:0.5]
      shade[0] = shade[0] - floor(shade[0]);    // Force into [0,1)
    }
    OC_REAL4 x = static_cast<OC_REAL4>((v.x+1)/2);
    shade[1] = (x<0 ? 0 : (x>1 ? 1 : x)); // Force into [0,1]
  }
  Vf_ColorQuantity GVS_yzangle("yz-angle",
                               {GetVecShade_yzangle,
                                Vf_ColorQuantity::EdgeType::Periodic});
  /// Instantiation registers name + {function, edgetype}

} // unnamed namespace end


#if DO_DEBUG
void GetVecShade_foobar
(const Nb_Vec3<OC_REAL4>& v,
 Vf_DisplayVector::SHADEVEC_T& shade)
{}

int main(int,char**)
{
  std::cout << "Hello! These are the functions I know:" << std::endl;
  Vf_ColorQuantity::PrintNames();
  std::cout << "Adding some more...\n";
  {
    Vf_ColorQuantity alpha("alpha",GetVecShade_foobar);
    Vf_ColorQuantity::PrintNames();
    std::cout << "---\n";
    Vf_ColorQuantity beta("beta",GetVecShade_foobar);
    Vf_ColorQuantity::PrintNames();
    std::cout << "---\n";
  }
  std::cout << "Destruction test...\n";
  Vf_ColorQuantity::PrintNames();

  std::cout << "Access test...\n";
  Vf_ColorQuantity::ColorQuantityAttributes attrib = Vf_ColorQuantity::GetAttributes("z");
  Vf_DisplayVector::SHADEVEC_T shade;
  (*(attrib.fcn))({1.2,3.4,5.6},shade);
  std::cout << "z-value: " << shade[0] << std::endl;

  std::cout << "Bye!" << std::endl;
  return 0;
}
#endif // DO_DEBUG
